from __future__ import annotations

import logging
from collections import defaultdict
from dataclasses import dataclass
from textwrap import dedent
from typing import Any, Callable, ClassVar, Optional, TypeVar

from typing_extensions import ParamSpec

from great_expectations.compatibility import docstring_parser
from great_expectations.compatibility.typing_extensions import override

logger = logging.getLogger(__name__)

WHITELISTED_TAG = "--Public API--"

P = ParamSpec("P")
T = TypeVar("T")
F = TypeVar("F", bound=Callable[..., Any])


def _remove_suffix(target: str, suffix: str) -> str:
    end_index = len(target) - len(suffix)
    if target.rfind(suffix) == end_index:
        return target[:end_index]
    return target


@dataclass(frozen=True)
class _PublicApiInfo:
    type: str
    name: str
    qualname: str
    module: Optional[str]


class _PublicApiIntrospector:
    _public_api: dict[str, list[_PublicApiInfo]] = {}

    # Only used for testing
    _class_registry: dict[str, set[str]] = defaultdict(set)
    _docstring_violations: set[str] = set()

    # This is a special key that is used to indicate that a class definition
    # is being added to the registry.
    CLASS_DEFINITION: ClassVar[str] = "<class_def>"

    @property
    def class_registry(self) -> dict[str, set[str]]:
        return self._class_registry

    @property
    def docstring_violations(self) -> set[str]:
        return self._docstring_violations

    def add(self, func: F) -> None:
        self._add_to_docstring_violations(func)
        self._add_to_class_registry(func)

        try:
            # We use an if statement instead of a ternary to work around
            # mypy's inability to type narrow inside a ternary.
            f: F
            if isinstance(func, classmethod):
                f = func.__func__
            else:
                f = func

            info = _PublicApiInfo(
                name=f.__name__,
                qualname=f.__qualname__,
                type=f.__class__.__name__,
                module=f.__module__ if hasattr(func, "__module__") else None,
            )
            if info.type not in self._public_api:
                self._public_api[info.type] = []
            self._public_api[info.type].append(info)
        except Exception:
            logger.exception(f"Could not add this function to the public API list: {func}")
            raise

    def _add_to_docstring_violations(self, func: F) -> None:
        name = f"{func.__module__}.{func.__qualname__}"
        if not func.__doc__ and name.startswith("great_expectations"):
            self._docstring_violations.add(name)

    def _add_to_class_registry(self, func: F) -> None:
        if isinstance(func, type):
            self._add_class_definition_to_registry(func)
        else:
            self._add_method_to_registry(func)

    def _add_class_definition_to_registry(self, cls: type) -> None:
        key = f"{cls.__module__}.{cls.__qualname__}"
        self._class_registry[key].add(self.CLASS_DEFINITION)

    def _add_method_to_registry(self, func: F) -> None:
        parts = func.__qualname__.split(".")
        METHOD_PARTS_LENGTH = 2
        if len(parts) == METHOD_PARTS_LENGTH:
            cls = parts[0]
            method = parts[1]
            key = f"{func.__module__}.{cls}"
            self._class_registry[key].add(method)
        elif len(parts) > METHOD_PARTS_LENGTH:
            # public_api interacts oddly with closures so we ignore
            # This is only present in DataSourceManager and its dynamic registry
            logger.info(
                "Skipping registering function %s because it is a closure",
                func.__qualname__,
            )
        else:
            # Standalone functions will have a length of 1
            logger.info(
                "Skipping registering function %s because it does not have a class",
                func.__qualname__,
            )

    @override
    def __str__(self) -> str:
        out = []
        for t in sorted(list(self._public_api.keys())):
            out.append(f"{t}")
            for info in sorted(self._public_api[t], key=lambda info: info.qualname):
                supporting_info = ""
                if info.name != info.qualname:
                    supporting_info = _remove_suffix(info.qualname, "." + info.name)
                elif info.module is not None:
                    supporting_info = info.module
                out.append(f"    {info.name}, {supporting_info}")
        return "\n".join(out)


public_api_introspector = _PublicApiIntrospector()


def public_api(func: F) -> F:
    """Add the public API tag for processing by the auto documentation generator.

    Used as a decorator:

        @public_api
        def my_method(some_argument):
            ...

    This tag is added at import time.
    """
    public_api_introspector.add(func)
    existing_docstring = func.__doc__ or ""
    func.__doc__ = WHITELISTED_TAG + existing_docstring
    return func


def deprecated_method_or_class(
    version: str,
    message: str = "",
) -> Callable[[F], F]:
    """Add a deprecation warning to the docstring of the decorated method or class.

    Used as a decorator:

        @deprecated_method_or_class(version="1.2.3", message="Optional message")
        def my_method(some_argument):
            ...

        or

        @deprecated_method_or_class(version="1.2.3", message="Optional message")
        class MyClass:
            ...

    Args:
        version: Version number when the method was deprecated.
        message: Optional deprecation message.
    """

    text = f".. deprecated:: {version}\n    {message}"

    def wrapper(func: F) -> F:
        """Wrapper method that accepts func, so we can modify the docstring."""
        return _add_text_to_function_docstring_after_summary(
            func=func,
            text=text,
        )

    return wrapper


def new_method_or_class(
    version: str,
    message: str = "",
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Add a version added note to the docstring of the decorated method or class.

    Used as a decorator:

        @new_method_or_class(version="1.2.3", message="Optional message")
        def my_method(some_argument):
            ...

        or

        @new_method_or_class(version="1.2.3", message="Optional message")
        class MyClass:
            ...

    Args:
        version: Version number when the method was added.
        message: Optional message.
    """

    text = f".. versionadded:: {version}\n    {message}"

    def wrapper(func: Callable[P, T]) -> Callable[P, T]:
        """Wrapper method that accepts func, so we can modify the docstring."""
        return _add_text_to_function_docstring_after_summary(
            func=func,
            text=text,
        )

    return wrapper


def deprecated_argument(
    argument_name: str,
    version: str,
    message: str = "",
) -> Callable[[F], F]:
    """Add an arg-specific deprecation warning to the decorated method or class.

    Used as a decorator:

        @deprecated_argument(argument_name="some_argument", version="1.2.3", message="Optional message")
        def my_method(some_argument):
            ...

        or

        @deprecated_argument(argument_name="some_argument", version="1.2.3", message="Optional message")
        class MyClass:
            ...

    If docstring_parser is not installed, this will not modify the docstring.

    Args:
        argument_name: Name of the argument to associate with the deprecation note.
        version: Version number when the method was deprecated.
        message: Optional deprecation message.
    """  # noqa: E501 # FIXME CoP

    text = f".. deprecated:: {version}\n    {message}"

    def wrapper(func: F) -> F:
        """Wrapper method that accepts func, so we can modify the docstring."""
        if not docstring_parser.docstring_parser:
            return func

        return _add_text_below_function_docstring_argument(
            func=func,
            argument_name=argument_name,
            text=text,
        )

    return wrapper


def new_argument(
    argument_name: str,
    version: str,
    message: str = "",
) -> Callable[[F], F]:
    """Add an arg-specific version added note to the decorated method or class.

    Used as a decorator:

        @new_argument(argument_name="some_argument", version="1.2.3", message="Optional message")
        def my_method(some_argument):
            ...

        or

        @new_argument(argument_name="some_argument", version="1.2.3", message="Optional message")
        class MyClass:
            ...

    If docstring_parser is not installed, this will not modify the docstring.

    Args:
        argument_name: Name of the argument to associate with the note.
        version: The version number to associate with the note.
        message: Optional message.
    """

    text = f".. versionadded:: {version}\n    {message}"

    def wrapper(func: F) -> F:
        """Wrapper method that accepts func, so we can modify the docstring."""
        if not docstring_parser.docstring_parser:
            return func

        return _add_text_below_function_docstring_argument(
            func=func,
            argument_name=argument_name,
            text=text,
        )

    return wrapper


def _add_text_to_function_docstring_after_summary(func: F, text: str) -> F:
    """Insert text into docstring, e.g. rst directive.

    Args:
        func: Add text to provided func docstring.
        text: String to add to the docstring, can be a rst directive e.g.:
            text = (
                ".. versionadded:: 1.2.3\n"
                "    Added in version 1.2.3\n"
            )

    Returns:
        func with modified docstring.
    """
    existing_docstring = func.__doc__ if func.__doc__ else ""
    split_docstring = existing_docstring.split("\n", 1)

    docstring = ""
    if len(split_docstring) == 2:  # noqa: PLR2004 # FIXME CoP
        short_description, docstring = split_docstring
        docstring = f"{short_description.strip()}\n\n{text}\n\n{dedent(docstring)}"
    elif len(split_docstring) == 1:
        short_description = split_docstring[0]
        docstring = f"{short_description.strip()}\n\n{text}\n"
    elif len(split_docstring) == 0:
        docstring = f"{text}\n"

    func.__doc__ = docstring

    return func


def _add_text_below_function_docstring_argument(
    func: F,
    argument_name: str,
    text: str,
) -> F:
    """Add text below specified docstring argument.

    Args:
        func: Callable[P, T]unction whose docstring will be modified.
        argument_name: Name of the argument to add text to its description.
        text: Text to add to the argument description.

    Returns:
        func with modified docstring.
    """
    existing_docstring = func.__doc__ if func.__doc__ else ""

    func.__doc__ = _add_text_below_string_docstring_argument(
        docstring=existing_docstring, argument_name=argument_name, text=text
    )

    return func


def _add_text_below_string_docstring_argument(docstring: str, argument_name: str, text: str) -> str:
    """Add text below an argument in a docstring.

    Note: Can be used for rst directives.

    Args:
        docstring: Docstring to modify.
        argument_name: Argument to place text below.
        text: Text to place below argument. Can be an rst directive.

    Returns:
        Modified docstring.
    """
    parsed_docstring = docstring_parser.docstring_parser.parse(
        text=docstring,
        style=docstring_parser.DocstringStyle.GOOGLE,
    )

    arg_list = list(param.arg_name for param in parsed_docstring.params)
    if argument_name not in arg_list:
        raise ValueError(f"Please specify an existing argument, you specified {argument_name}.")  # noqa: TRY003 # FIXME CoP

    for param in parsed_docstring.params:
        if param.arg_name == argument_name:
            if param.description is None:
                param.description = text
            else:
                param.description += "\n\n" + text + "\n"

    # Returns: includes an additional ":\n" that we need to strip out.
    if parsed_docstring.returns:
        if parsed_docstring.returns.description:
            parsed_docstring.returns.description = parsed_docstring.returns.description.strip(":\n")

    # RenderingStyle.EXPANDED used to make sure any line breaks before and
    # after the added text are included (for Sphinx html rendering).
    composed_docstring = docstring_parser.docstring_parser.compose(
        docstring=parsed_docstring,
        style=docstring_parser.DocstringStyle.GOOGLE,
        rendering_style=docstring_parser.docstring_parser.RenderingStyle.EXPANDED,
    )

    return composed_docstring



import logging
import pprint
from dataclasses import dataclass
from typing import Callable, Generator, Mapping, Optional, Sequence, TypeVar, Union
from uuid import UUID

import pandas as pd
import pytest
from _pytest.mark import MarkDecorator

import great_expectations as gx
from great_expectations.compatibility.typing_extensions import override
from great_expectations.data_context.data_context.context_factory import set_context
from great_expectations.datasource.fluent.interfaces import Batch, DataAsset
from tests.integration.sql_session_manager import SessionSQLEngineManager
from tests.integration.test_utils.data_source_config import DataSourceTestConfig
from tests.integration.test_utils.data_source_config.base import (
    BatchTestSetup,
    dict_to_tuple,
    hash_data_frame,
)
from tests.integration.test_utils.data_source_config.sql import (
    SQLBatchTestSetup,
)

logger = logging.getLogger(__name__)

_F = TypeVar("_F", bound=Callable)


@dataclass(frozen=True)
class TestConfig:
    data_source_config: DataSourceTestConfig
    data: pd.DataFrame
    extra_data: Mapping[str, pd.DataFrame]
    secondary_source_config: Union[DataSourceTestConfig, None] = None
    secondary_data: Union[pd.DataFrame, None] = None

    @override
    def __hash__(self) -> int:
        if self.secondary_data is None and self.secondary_source_config is not None:
            raise ValueError(
                "secondary_data cannot be None when secondary_source_config is provided"
            )
        return hash(
            (
                self.__class__,
                hash_data_frame(self.data),
                dict_to_tuple(
                    {k: hash_data_frame(self.extra_data[k]) for k in sorted(self.extra_data)}
                ),
                hash_data_frame(self.secondary_data) if self.secondary_data is not None else None,
            )
        )

    @override
    def __eq__(self, value: object) -> bool:
        # We need to implement this ourselves to call `.equals` on dataframes.`
        if not isinstance(value, TestConfig):
            return False
        return all(
            [
                self.data_source_config == value.data_source_config,
                self.data.equals(value.data),
                self.extra_data.keys() == value.extra_data.keys(),
                all(self.extra_data[k].equals(value.extra_data[k]) for k in self.extra_data),
                self.secondary_source_config == value.secondary_source_config,
                (self.secondary_data is None and value.secondary_data is None)
                or (
                    self.secondary_data is not None
                    and value.secondary_data is not None
                    and self.secondary_data.equals(value.secondary_data)
                ),
            ]
        )


def parameterize_batch_for_data_sources(
    data_source_configs: Sequence[DataSourceTestConfig],
    data: pd.DataFrame,
    extra_data: Optional[Mapping[str, pd.DataFrame]] = None,
) -> Callable[[_F], _F]:
    """Test decorator that parametrizes a test function with batches for various data sources.
    This injects a `batch_for_datasource` parameter into the test function for each data source
    type.

    Args:
        data_source_configs: The data source configurations to test.
        data: Data to load into the asset
        extra_data: Mapping of {asset_label: data} to load into other assets. Only relevant for SQL
                    multi-table expectations. NOTE: This is NOT the table name. The label is used to
                    correlate the data with the types passed to
                    DataSourceTestConfig.extra_column_types.


    example use:
        @parameterize_batch_for_data_sources(
            data_source_configs=[DataSourceType.FOO, DataSourceType.BAR],
            data=pd.DataFrame{"col_name": [1, 2]},
            # description="test_stuff",
        )
        def test_stuff(batch_for_datasource) -> None:
            ...
    """
    if len(data_source_configs) == 0:
        raise ValueError("At least one data source configuration must be provided.")

    def decorator(func: _F) -> _F:
        pytest_params = [
            pytest.param(
                TestConfig(
                    data_source_config=config,
                    data=data,
                    extra_data=extra_data or {},
                ),
                id=config.test_id,
                marks=[config.pytest_mark],
            )
            for config in data_source_configs
        ]
        parameterize_decorator = pytest.mark.parametrize(
            _batch_setup_for_datasource.__name__,
            pytest_params,
            indirect=True,
        )
        return parameterize_decorator(func)

    return decorator


# NOTE on performance setup/teardown:
# When we get equivalent TestConfigs, we only instantiate one BatchTestSetup for all of them, and
# only perform its setup/teardown once. batch_for_datasource instantiate the BatchTestSetup
# immediately before the first test that needs it and store it in cached_test_configs. Subsequent
# tests that use the same TestConfig will reuse the same BatchTestSetup. At the end of the test
# session, _cleanup will clean up all the BatchTestSetups.


@pytest.fixture(scope="session")
def _cached_test_configs() -> dict[TestConfig, BatchTestSetup]:
    """Fixture to hold cached test configurations across tests."""
    cached_test_configs: dict[TestConfig, BatchTestSetup] = {}

    return cached_test_configs


@pytest.fixture(scope="session")
def _cached_secondary_test_configs() -> dict[UUID, BatchTestSetup]:
    """Fixture to hold secondary test configurations across tests."""
    cached_test_configs: dict[UUID, BatchTestSetup] = {}
    return cached_test_configs


@pytest.fixture(scope="session")
def _cleanup(
    _cached_test_configs: Mapping[TestConfig, BatchTestSetup],
    _cached_secondary_test_configs: Mapping[TestConfig, BatchTestSetup],
    # While not explicitly used, we want the session_sql_engine_manager to
    # be torn down after we clean up. Adding it as a dependeny will ensure
    # this.
    session_sql_engine_manager: SessionSQLEngineManager,
) -> Generator[None, None, None]:
    """Fixture to do all teardown at the end of the test session."""
    yield
    for batch_setup in _cached_test_configs.values():
        batch_setup.teardown()
    for batch_setup in _cached_secondary_test_configs.values():
        batch_setup.teardown()


@pytest.fixture
def _batch_setup_for_datasource(
    request: pytest.FixtureRequest,
    _cached_test_configs: dict[TestConfig, BatchTestSetup],
    _cached_secondary_test_configs: dict[UUID, BatchTestSetup],
    session_sql_engine_manager: SessionSQLEngineManager,
    # _cleanup is not called directly. It is a session scoped fixture
    # which will cleanup created db resources such as schemas.
    _cleanup: Callable[[], None],
) -> Generator[BatchTestSetup, None, None]:
    """Fixture that yields a BatchSetup for a specific data source type.
    This must be used in conjunction with `indirect=True` to defer execution
    """
    config = request.param
    assert isinstance(config, TestConfig)

    if config not in _cached_test_configs:
        batch_setup = config.data_source_config.create_batch_setup(
            request=request,
            data=config.data,
            extra_data=config.extra_data,
            context=gx.get_context(mode="ephemeral"),
            engine_manager=session_sql_engine_manager,
        )
        _cached_test_configs[config] = batch_setup
        batch_setup.setup()
        if config.secondary_source_config:
            assert config.secondary_data is not None, (
                "Secondary data is required when secondary config is provided."
            )
            secondary_batch_setup = config.secondary_source_config.create_batch_setup(
                request=request,
                data=config.secondary_data,
                extra_data={},
                context=batch_setup.context,
                engine_manager=session_sql_engine_manager,
            )
            _cached_secondary_test_configs[batch_setup.id] = secondary_batch_setup
            secondary_batch_setup.setup()

    yield _cached_test_configs[config]


@pytest.fixture
def batch_for_datasource(
    _batch_setup_for_datasource: BatchTestSetup,
) -> Generator[Batch, None, None]:
    """Fixture that yields a batch for a specific data source type.
    This must be used in conjunction with `indirect=True` to defer execution
    """
    set_context(_batch_setup_for_datasource.context)
    yield _batch_setup_for_datasource.make_batch()


@pytest.fixture
def asset_for_datasource(
    _batch_setup_for_datasource: BatchTestSetup,
) -> Generator[DataAsset, None, None]:
    """Fixture that yields an asset for a specific data source type.
    This must be used in conjunction with `indirect=True` to defer execution
    """
    set_context(_batch_setup_for_datasource.context)
    yield _batch_setup_for_datasource.make_asset()


@dataclass(frozen=True)
class MultiSourceBatch:
    base_batch: Batch
    comparison_data_source_name: str
    comparison_table_name: str


@pytest.fixture
def multi_source_batch(
    _batch_setup_for_datasource: BatchTestSetup,
    _cached_secondary_test_configs: dict[UUID, BatchTestSetup],
) -> Generator[MultiSourceBatch, None, None]:
    """Fixture that sets up multiple sources in a single data context."""
    secondary_batch_setup = _cached_secondary_test_configs[_batch_setup_for_datasource.id]
    assert isinstance(secondary_batch_setup, SQLBatchTestSetup), (
        "MultiSourceBatch requires SQLBatchTestSetup"
    )
    # we need a data source, so we use the make_asset API:
    set_context(_batch_setup_for_datasource.context)
    secondary_asset = secondary_batch_setup.make_asset()
    yield MultiSourceBatch(
        base_batch=_batch_setup_for_datasource.make_batch(),
        comparison_data_source_name=secondary_asset.datasource.name,
        comparison_table_name=secondary_batch_setup.table_name,
    )


@pytest.fixture
def extra_table_names_for_datasource(
    _batch_setup_for_datasource: BatchTestSetup,
) -> Generator[Mapping[str, str], None, None]:
    """Fixture that yields extra table names"""
    assert isinstance(_batch_setup_for_datasource, SQLBatchTestSetup)
    yield {key: t.name for key, t in _batch_setup_for_datasource.extra_table_data.items()}


@pytest.fixture(scope="session")
def _base_to_comparison_map() -> Mapping[UUID, UUID]:
    """Get a comparison BatchTestSetup ID by its base BatchTestSetup ID."""
    return {}


@dataclass(frozen=True)
class MultiSourceTestConfig:
    comparison: DataSourceTestConfig
    base: DataSourceTestConfig


def multi_source_batch_setup(
    multi_source_test_configs: list[MultiSourceTestConfig],
    base_data: pd.DataFrame,
    comparison_data: pd.DataFrame,
) -> Callable[[_F], _F]:
    def decorator(func: _F) -> _F:
        pytest_params = []
        for multi_source_test_config in multi_source_test_configs:
            pytest_params.append(
                pytest.param(
                    TestConfig(
                        data_source_config=multi_source_test_config.base,
                        data=base_data,
                        extra_data={},
                        secondary_source_config=multi_source_test_config.comparison,
                        secondary_data=comparison_data,
                    ),
                    id=f"{multi_source_test_config.comparison.test_id}->{multi_source_test_config.base.test_id}",
                    marks=_get_multi_source_marks(multi_source_test_config),
                )
            )
        parameterize_decorator = pytest.mark.parametrize(
            _batch_setup_for_datasource.__name__,
            pytest_params,
            indirect=True,
        )
        return parameterize_decorator(func)

    return decorator


def _get_multi_source_marks(multi_source_test_config: MultiSourceTestConfig) -> list[MarkDecorator]:
    if multi_source_test_config.base.pytest_mark == multi_source_test_config.comparison.pytest_mark:
        return [multi_source_test_config.base.pytest_mark]
    # our test setup restricts us to testing a single backend at a time.
    # sqlite doesn't require any extra setup, so it's an exception.
    marks = [
        mark
        for mark in [
            multi_source_test_config.comparison.pytest_mark,
            multi_source_test_config.base.pytest_mark,
        ]
        if mark != pytest.mark.sqlite
    ]
    if len(marks) == 1:
        return marks
    elif not marks:
        return [pytest.mark.sqlite]
    else:
        raise ValueError(
            "MultiSourceBatch tests must either use the same backend or include sqlite."
        )


@pytest.fixture(scope="session")
def session_sql_engine_manager():
    logger.info("SessionSqlEngineManager: Starting setup.")
    manager = SessionSQLEngineManager()
    yield manager

    logger.info("SessionSqlEngineManager: Starting teardown.")
    pre_cleanup_stats = manager.get_all_pool_statistics()
    # We temporarily log a warning so we can see this in the pytest output without turning on info
    # logging across the whole test run.
    logger.warning(
        "SessionSqlEngineManager: Pool statistics before explicit cleanup:\n"
        f"{pprint.pformat(pre_cleanup_stats)}"
    )
    # Check for any immediately obvious issues before cleanup
    for key, stat in pre_cleanup_stats.items():
        if "error" not in stat and stat.get("checked_out", 0) > 0:
            logger.warning(
                f"SessionSqlEngineManager: Engine {key} has {stat['checked_out']} connections "
                "still checked out BEFORE manager disposal."
            )
    manager.dispose_all_engines()
    logger.info("SessionSqlEngineManager: All engines disposed by manager.")
    assert not manager._engine_cache, "Engine cache should be empty after dispose_all_engines."
    logger.info("SessionSqlEngineManager: Teardown complete.")



from __future__ import annotations

from great_expectations.compatibility.not_imported import NotImported

SQLALCHEMY_BIGQUERY_NOT_IMPORTED = NotImported(
    "sqlalchemy-bigquery is not installed, please 'pip install sqlalchemy-bigquery'"
)
_BIGQUERY_MODULE_NAME = "sqlalchemy_bigquery"
BIGQUERY_GEO_SUPPORT = False

bigquery_types_tuple = None


try:
    import sqlalchemy_bigquery
except (ImportError, AttributeError):
    sqlalchemy_bigquery = SQLALCHEMY_BIGQUERY_NOT_IMPORTED

try:
    from sqlalchemy_bigquery import INTEGER
except (ImportError, AttributeError):
    INTEGER = SQLALCHEMY_BIGQUERY_NOT_IMPORTED

try:
    from sqlalchemy_bigquery import NUMERIC
except (ImportError, AttributeError):
    NUMERIC = SQLALCHEMY_BIGQUERY_NOT_IMPORTED

try:
    from sqlalchemy_bigquery import STRING
except (ImportError, AttributeError):
    STRING = SQLALCHEMY_BIGQUERY_NOT_IMPORTED

try:
    from sqlalchemy_bigquery import BIGNUMERIC
except (ImportError, AttributeError):
    BIGNUMERIC = SQLALCHEMY_BIGQUERY_NOT_IMPORTED

try:
    from sqlalchemy_bigquery import BYTES
except (ImportError, AttributeError):
    BYTES = SQLALCHEMY_BIGQUERY_NOT_IMPORTED


try:
    from sqlalchemy_bigquery import BOOL
except (ImportError, AttributeError):
    BOOL = SQLALCHEMY_BIGQUERY_NOT_IMPORTED


try:
    from sqlalchemy_bigquery import BOOLEAN
except (ImportError, AttributeError):
    BOOLEAN = SQLALCHEMY_BIGQUERY_NOT_IMPORTED


try:
    from sqlalchemy_bigquery import TIMESTAMP
except (ImportError, AttributeError):
    TIMESTAMP = SQLALCHEMY_BIGQUERY_NOT_IMPORTED


try:
    from sqlalchemy_bigquery import TIME
except (ImportError, AttributeError):
    TIME = SQLALCHEMY_BIGQUERY_NOT_IMPORTED

try:
    from sqlalchemy_bigquery import FLOAT
except (ImportError, AttributeError):
    FLOAT = SQLALCHEMY_BIGQUERY_NOT_IMPORTED


try:
    from sqlalchemy_bigquery import DATE
except (ImportError, AttributeError):
    DATE = SQLALCHEMY_BIGQUERY_NOT_IMPORTED


try:
    from sqlalchemy_bigquery import DATETIME
except (ImportError, AttributeError):
    DATETIME = SQLALCHEMY_BIGQUERY_NOT_IMPORTED


class BIGQUERY_TYPES:
    """Namespace for Bigquery dialect types"""

    INTEGER = INTEGER
    NUMERIC = NUMERIC
    STRING = STRING
    BIGNUMERIC = BIGNUMERIC
    BYTES = BYTES
    BOOL = BOOL
    BOOLEAN = BOOLEAN
    TIMESTAMP = TIMESTAMP
    TIME = TIME
    FLOAT = FLOAT
    DATE = DATE
    DATETIME = DATETIME


try:
    from sqlalchemy_bigquery import GEOGRAPHY

    BIGQUERY_GEO_SUPPORT = True
except (ImportError, AttributeError):
    GEOGRAPHY = SQLALCHEMY_BIGQUERY_NOT_IMPORTED

try:
    from sqlalchemy_bigquery import parse_url
except (ImportError, AttributeError):
    parse_url = SQLALCHEMY_BIGQUERY_NOT_IMPORTED



from datetime import datetime, timezone

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations import get_context
from great_expectations.compatibility.sqlalchemy import sqltypes
from tests.integration.test_utils.data_source_config import PostgreSQLDatasourceTestConfig
from tests.integration.test_utils.data_source_config.postgres import PostgresBatchTestSetup

pytestmark = pytest.mark.postgresql


class TestPostgresqlDataTypes:
    """This set of tests ensures that we can run expectations against every data
    type supported by Postgres.

    https://www.postgresql.org/docs/current/datatype.html
    """

    BOOL_COL_NAME = "my_bool"
    DATE_COL_NAME = "my_date"
    NUMERIC_COL_NAME = "my_number"
    STRING_COL_NAME = "my_string"

    DATA_FRAME = pd.DataFrame(
        {
            BOOL_COL_NAME: [True, False, True, False],
            DATE_COL_NAME: [
                datetime(2021, 1, 1, tzinfo=timezone.utc).date(),
                datetime(2021, 1, 2, tzinfo=timezone.utc).date(),
                datetime(2021, 1, 3, tzinfo=timezone.utc).date(),
                datetime(2021, 1, 4, tzinfo=timezone.utc).date(),
            ],
            NUMERIC_COL_NAME: [1, 2, 3, 4],
            STRING_COL_NAME: ["a", "b", "c", "d"],
        }
    )

    def test_boolean(self):
        batch_setup = PostgresBatchTestSetup(
            config=PostgreSQLDatasourceTestConfig(
                column_types={self.BOOL_COL_NAME: sqltypes.BOOLEAN}
            ),
            data=self.DATA_FRAME,
            extra_data={},
            context=get_context(mode="ephemeral"),
        )

        with batch_setup.batch_test_context() as batch:
            result = batch.validate(
                gxe.ExpectColumnValuesToBeInSet(
                    column=self.BOOL_COL_NAME,
                    value_set=[True, False],
                )
            )
        assert result.success

    @pytest.mark.parametrize(
        "col_type",
        [
            sqltypes.TIMESTAMP(timezone=False),
            sqltypes.TIMESTAMP(timezone=True),
            sqltypes.DATE,
        ],
    )
    def test_dates(self, col_type):
        batch_setup = PostgresBatchTestSetup(
            config=PostgreSQLDatasourceTestConfig(column_types={self.DATE_COL_NAME: col_type}),
            data=self.DATA_FRAME,
            extra_data={},
            context=get_context(mode="ephemeral"),
        )

        with batch_setup.batch_test_context() as batch:
            result = batch.validate(
                gxe.ExpectColumnValuesToBeBetween(
                    column=self.DATE_COL_NAME,
                    min_value=datetime(2020, 1, 1, tzinfo=timezone.utc).date(),
                    max_value=datetime(2022, 1, 1, tzinfo=timezone.utc).date(),
                )
            )
        assert result.success

    @pytest.mark.parametrize(
        "col_type",
        [
            sqltypes.SMALLINT,
            sqltypes.INT,
            sqltypes.BIGINT,
            sqltypes.DECIMAL,
            sqltypes.FLOAT,
            sqltypes.NUMERIC,
            sqltypes.REAL,
        ],
    )
    def test_numbers(self, col_type):
        batch_setup = PostgresBatchTestSetup(
            config=PostgreSQLDatasourceTestConfig(column_types={self.NUMERIC_COL_NAME: col_type}),
            data=self.DATA_FRAME,
            extra_data={},
            context=get_context(mode="ephemeral"),
        )

        with batch_setup.batch_test_context() as batch:
            result = batch.validate(
                gxe.ExpectColumnSumToBeBetween(
                    column=self.NUMERIC_COL_NAME,
                    min_value=9,
                    max_value=11,
                )
            )
        assert result.success

    @pytest.mark.parametrize(
        "col_type",
        [
            sqltypes.CHAR,
            sqltypes.TEXT,
        ],
    )
    def test_strings(self, col_type):
        batch_setup = PostgresBatchTestSetup(
            config=PostgreSQLDatasourceTestConfig(column_types={self.STRING_COL_NAME: col_type}),
            data=self.DATA_FRAME,
            extra_data={},
            context=get_context(mode="ephemeral"),
        )

        with batch_setup.batch_test_context() as batch:
            result = batch.validate(
                gxe.ExpectColumnValuesToBeInSet(
                    column=self.STRING_COL_NAME, value_set=["a", "b", "c", "d"]
                )
            )
        assert result.success



from __future__ import annotations

import copy
import logging
from abc import ABC, abstractmethod
from pprint import pformat as pf
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Generic,
    List,
    Mapping,
    Optional,
    Set,
)

import great_expectations.exceptions as gx_exceptions
from great_expectations.compatibility import pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.batch_definition import PartitionerT
from great_expectations.datasource.fluent.batch_request import (
    BatchRequest,
)
from great_expectations.datasource.fluent.data_connector import (
    FilePathDataConnector,  # noqa: TC001  # pydantic uses type at runtime
)
from great_expectations.datasource.fluent.interfaces import (
    Batch,
    DataAsset,
    DatasourceT,
    PartitionerSortingProtocol,
    TestConnectionError,
)
from great_expectations.exceptions.exceptions import NoAvailableBatchesError

if TYPE_CHECKING:
    from great_expectations.alias_types import PathStr
    from great_expectations.core.batch import LegacyBatchDefinition
    from great_expectations.execution_engine import (
        PandasExecutionEngine,
        SparkDFExecutionEngine,
    )

logger = logging.getLogger(__name__)


class PathDataAsset(DataAsset, Generic[DatasourceT, PartitionerT], ABC):
    _EXCLUDE_FROM_READER_OPTIONS: ClassVar[Set[str]] = {
        "batch_definitions",
        "type",
        "name",
        "order_by",
        "batch_metadata",
        "batching_regex",  # file_path argument
        "kwargs",  # kwargs need to be unpacked and passed separately
        "connect_options",
        "id",
    }

    # General file-path DataAsset pertaining attributes.

    connect_options: Mapping = pydantic.Field(
        default_factory=dict,
        description="Optional filesystem specific advanced parameters for connecting to data assets",  # noqa: E501 # FIXME CoP
    )

    # `_data_connector`` should be set inside `_build_data_connector()`
    _data_connector: FilePathDataConnector = pydantic.PrivateAttr()
    # more specific `_test_connection_error_message` can be set inside `_build_data_connector()`
    _test_connection_error_message: str = pydantic.PrivateAttr("Could not connect to your asset")

    class Config:
        """
        Need to allow extra fields for the base type because pydantic will first create
        an instance of `PathDataAsset` before we select and create the more specific
        asset subtype.
        Each specific subtype should `forbid` extra fields.
        """

        extra = pydantic.Extra.allow

    @override
    def get_batch_parameters_keys(
        self,
        partitioner: Optional[PartitionerT] = None,
    ) -> tuple[str, ...]:
        raise NotImplementedError

    @override
    def _validate_batch_request(self, batch_request: BatchRequest) -> None:
        """Validates the batch_request has the correct form.

        Args:
            batch_request: A batch request object to be validated.
        """
        if not (
            batch_request.datasource_name == self.datasource.name
            and batch_request.data_asset_name == self.name
            and self._batch_parameters_are_valid(
                options=batch_request.options, partitioner=batch_request.partitioner
            )
        ):
            valid_options = self.get_batch_parameters_keys(partitioner=batch_request.partitioner)
            options = dict.fromkeys(valid_options)
            expect_batch_request_form = BatchRequest(
                datasource_name=self.datasource.name,
                data_asset_name=self.name,
                options=options,
                batch_slice=batch_request._batch_slice_input,  # type: ignore[attr-defined] # FIXME CoP
                partitioner=batch_request.partitioner,
            )
            raise gx_exceptions.InvalidBatchRequestError(  # noqa: TRY003 # FIXME CoP
                "BatchRequest should have form:\n"
                f"{pf(expect_batch_request_form.dict())}\n"
                f"but actually has form:\n{pf(batch_request.dict())}\n"
            )

    @override
    def get_batch_identifiers_list(self, batch_request: BatchRequest) -> List[dict]:
        batch_definition_list = self._get_batch_definition_list(batch_request)
        batch_identifiers_list: List[dict] = [
            batch_definition_list.batch_identifiers
            for batch_definition_list in batch_definition_list
        ]
        if sortable_partitioner := self._get_sortable_partitioner(batch_request.partitioner):
            batch_identifiers_list = self.sort_batch_identifiers_list(
                batch_identifiers_list, sortable_partitioner
            )

        return batch_identifiers_list

    @override
    def get_batch(self, batch_request: BatchRequest) -> Batch:
        """Get a batch from the data asset using a batch request."""
        self._validate_batch_request(batch_request)

        execution_engine: PandasExecutionEngine | SparkDFExecutionEngine = (
            self.datasource.get_execution_engine()
        )

        batch_definitions = self._get_batch_definition_list(batch_request)
        if not batch_definitions:
            raise NoAvailableBatchesError()

        # Pick the last, which most likely corresponds to the most recent, batch in the list
        if sortable_partitioner := self._get_sortable_partitioner(batch_request.partitioner):
            batch_definitions = self.sort_legacy_batch_definitions(
                batch_definitions,
                sortable_partitioner,
            )
        batch_definition = batch_definitions[-1]

        batch_spec = self._data_connector.build_batch_spec(batch_definition=batch_definition)
        batch_spec_options = self._batch_spec_options_from_batch_request(batch_request)
        batch_spec.update(batch_spec_options)

        data, markers = execution_engine.get_batch_data_and_markers(batch_spec=batch_spec)

        fully_specified_batch_request = copy.deepcopy(batch_request)
        fully_specified_batch_request.options.update(batch_definition.batch_identifiers)
        batch_metadata = self._get_batch_metadata_from_batch_request(
            batch_request=fully_specified_batch_request
        )

        return Batch(
            datasource=self.datasource,
            data_asset=self,
            batch_request=fully_specified_batch_request,
            data=data,
            metadata=batch_metadata,
            batch_markers=markers,
            batch_spec=batch_spec,
            batch_definition=batch_definition,
        )

    def _get_batch_definition_list(
        self, batch_request: BatchRequest
    ) -> list[LegacyBatchDefinition]:
        """Generate a batch definition list from a given batch request.

        Args:
            batch_request: Batch request used to generate batch definitions.

        Returns:
            List of batch definitions.
        """
        raise NotImplementedError

    @override
    def test_connection(self) -> None:
        """Test the connection for the DataAsset.

        Raises:
            TestConnectionError: If the connection test fails.
        """
        try:
            if self._data_connector.test_connection():
                return None
        except Exception as e:
            raise TestConnectionError(  # noqa: TRY003 # FIXME CoP
                f"Could not connect to asset using {type(self._data_connector).__name__}: Got {type(e).__name__}"  # noqa: E501 # FIXME CoP
            ) from e
        raise TestConnectionError(self._test_connection_error_message)

    def _batch_spec_options_from_batch_request(self, batch_request: BatchRequest) -> dict:
        """Build a set of options for use in a batch spec from a batch request.

        Args:
            batch_request: Batch request to use to generate options.

        Returns:
            Dictionary containing batch spec options.
        """
        raise NotImplementedError

    def get_whole_directory_path_override(
        self,
    ) -> PathStr | None:
        """If present, override DataConnector behavior in order to
        treat an entire directory as a single Asset.
        """
        # todo: refactor data connector instantiation so this isn't necessary
        raise NotImplementedError

    def _get_reader_method(self) -> str:
        # subtypes must define a reader method
        raise NotImplementedError

    def _get_reader_options_include(self) -> set[str]:
        # subtypes control how reader options get serialized
        raise NotImplementedError

    @abstractmethod
    def _get_sortable_partitioner(
        self, partitioner: Optional[PartitionerT]
    ) -> Optional[PartitionerSortingProtocol]:
        # allow subclasses to determine sorting configuration.
        raise NotImplementedError



from typing import TYPE_CHECKING, Any, Generic, TypeVar, Union

import pandas as pd

from great_expectations.compatibility.pydantic import BaseModel, GenericModel
from great_expectations.validator.metric_configuration import (
    MetricConfigurationID,
)

if TYPE_CHECKING:
    from great_expectations.compatibility.pyspark import pyspark
    from great_expectations.compatibility.sqlalchemy import BinaryExpression

_MetricResultValue = TypeVar("_MetricResultValue")


class MetricResult(GenericModel, Generic[_MetricResultValue]):
    id: MetricConfigurationID
    value: _MetricResultValue

    class Config:
        arbitrary_types_allowed = True


class MetricErrorResultValue(BaseModel):
    exception_message: str
    exception_traceback: Union[str, None] = None


class MetricErrorResult(MetricResult[MetricErrorResultValue]): ...


class ConditionValuesValueError(ValueError):
    def __init__(self, actual_type: type) -> None:
        super().__init__(
            "Value must be one of: "
            "pandas.Series, pyspark.sql.Column, or sqlalchemy.BinaryExpression. "
            f"Got type `{actual_type}` instead."
        )


class ConditionValues(MetricResult[Union[pd.Series, "pyspark.sql.Column", "BinaryExpression"]]):
    @classmethod
    def validate_value_type(cls, value):
        if isinstance(value, pd.Series):
            return value

        try:
            from great_expectations.compatibility.pyspark import pyspark

            if isinstance(value, pyspark.sql.Column):
                return value
        except (ImportError, AttributeError):
            pass

        try:
            from great_expectations.compatibility.sqlalchemy import BinaryExpression

            if isinstance(value, BinaryExpression):
                return value
        except (ImportError, AttributeError):
            pass

        raise ConditionValuesValueError(type(value))

    @classmethod
    def __get_validators__(cls):
        yield cls.validate_value_type


class TableColumnsResult(MetricResult[list[str]]): ...


class ColumnType(BaseModel):
    class Config:
        extra = "allow"  # some backends return extra values

    name: str
    type: str


class TableColumnTypesResult(MetricResult[list[ColumnType]]): ...


class UnexpectedCountResult(MetricResult[int]): ...


class UnexpectedValuesResult(MetricResult[list[Any]]): ...



from __future__ import annotations

import logging
from functools import reduce

from great_expectations.compatibility.pyspark import functions as F
from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.map_metric_provider import (
    MulticolumnMapMetricProvider,
)
from great_expectations.expectations.metrics.map_metric_provider.multicolumn_condition_partial import (  # noqa: E501 # FIXME CoP
    multicolumn_condition_partial,
)

logger = logging.getLogger(__name__)


class SelectColumnValuesUniqueWithinRecord(MulticolumnMapMetricProvider):
    condition_metric_name = "select_column_values.unique.within_record"
    condition_domain_keys = (
        "batch_id",
        "table",
        "column_list",
        "row_condition",
        "condition_parser",
        "ignore_row_if",
    )

    @multicolumn_condition_partial(engine=PandasExecutionEngine)
    def _pandas(cls, column_list, **kwargs):
        num_columns = len(column_list.columns)
        row_wise_cond = column_list.nunique(dropna=False, axis=1) >= num_columns
        return row_wise_cond

    @multicolumn_condition_partial(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(cls, column_list, **kwargs):
        """
        The present approach relies on an inefficient query condition construction implementation, whose computational
        cost is O(num_columns^2).  However, until a more efficient implementation compatible with SQLAlchemy is
        available, this is the only feasible mechanism under the current architecture, where map metric providers must
        return a condition.  Nevertheless, SQL query length limit is 1GB (sufficient for most practical scenarios).
        """  # noqa: E501 # FIXME CoP
        num_columns = len(column_list)

        # An arbitrary "num_columns" value used for issuing an explanatory message as a warning.
        if num_columns > 100:  # noqa: PLR2004 # FIXME CoP
            logger.warning(
                f"""Batch data with {num_columns} columns is detected.  Computing the "{cls.condition_metric_name}" \
metric for wide tables using SQLAlchemy leads to long WHERE clauses for the underlying database engine to process.
"""  # noqa: E501 # FIXME CoP
            )

        conditions = sa.or_(
            *(
                sa.or_(
                    column_list[idx_src] == column_list[idx_dest],
                    sa.and_(
                        column_list[idx_src] == None,  # noqa: E711 # FIXME CoP
                        column_list[idx_dest] == None,  # noqa: E711 # FIXME CoP
                    ),
                )
                for idx_src in range(num_columns - 1)
                for idx_dest in range(idx_src + 1, num_columns)
            )
        )
        row_wise_cond = sa.not_(conditions)
        return row_wise_cond

    @multicolumn_condition_partial(engine=SparkDFExecutionEngine)
    def _spark(cls, column_list, **kwargs):
        column_names = column_list.columns
        num_columns = len(column_names)

        conditions = []
        for idx_src in range(num_columns - 1):
            for idx_dest in range(idx_src + 1, num_columns):
                conditions.append(
                    F.col(column_names[idx_src]).eqNullSafe(F.col(column_names[idx_dest]))
                )

        row_wise_cond = ~reduce(lambda a, b: a | b, conditions)
        return row_wise_cond



import pandas as pd

from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.fluent.conftest import (
    TEST_TABLE_NAME,
    _run_checkpoint_test,
    _run_column_expectation_test,
)
from tests.integration.test_utils.data_source_config import DatabricksDatasourceTestConfig


class TestDatabricksTableIdentifiers:
    @parameterize_batch_for_data_sources(
        data_source_configs=[
            DatabricksDatasourceTestConfig(table_name=TEST_TABLE_NAME.lower()),
        ],
        data=pd.DataFrame({"test_column": [1, 2, 3]}),
    )
    def test_unquoted_lower(self, batch_for_datasource):
        """Test Databricks with unquoted lower case table name"""
        _run_checkpoint_test(batch_for_datasource, "databricks")

    @parameterize_batch_for_data_sources(
        data_source_configs=[
            DatabricksDatasourceTestConfig(table_name=f"`{TEST_TABLE_NAME.lower()}`"),
        ],
        data=pd.DataFrame({"test_column": [1, 2, 3]}),
    )
    def test_quoted_lower(self, batch_for_datasource):
        """Test Databricks with quoted lower case table name"""
        _run_checkpoint_test(batch_for_datasource, "databricks")

    @parameterize_batch_for_data_sources(
        data_source_configs=[
            DatabricksDatasourceTestConfig(table_name=TEST_TABLE_NAME.upper()),
        ],
        data=pd.DataFrame({"test_column": [1, 2, 3]}),
    )
    def test_unquoted_upper(self, batch_for_datasource):
        """Test Databricks with unquoted upper case table name"""
        _run_checkpoint_test(batch_for_datasource, "databricks")

    @parameterize_batch_for_data_sources(
        data_source_configs=[
            DatabricksDatasourceTestConfig(table_name=f"`{TEST_TABLE_NAME.upper()}`"),
        ],
        data=pd.DataFrame({"test_column": [1, 2, 3]}),
    )
    def test_quoted_upper(self, batch_for_datasource):
        """Test Databricks with quoted upper case table name"""
        _run_checkpoint_test(batch_for_datasource, "databricks")

    @parameterize_batch_for_data_sources(
        data_source_configs=[
            DatabricksDatasourceTestConfig(table_name=f"`{TEST_TABLE_NAME.title()}`"),
        ],
        data=pd.DataFrame({"test_column": [1, 2, 3]}),
    )
    def test_quoted_mixed(self, batch_for_datasource):
        """Test Databricks with quoted mixed case table name"""
        _run_checkpoint_test(batch_for_datasource, "databricks")

    @parameterize_batch_for_data_sources(
        data_source_configs=[
            DatabricksDatasourceTestConfig(table_name=TEST_TABLE_NAME.title()),
        ],
        data=pd.DataFrame({"test_column": [1, 2, 3]}),
    )
    def test_unquoted_mixed(self, batch_for_datasource):
        """Test Databricks with unquoted mixed case table name"""
        _run_checkpoint_test(batch_for_datasource, "databricks")


class TestDatabricksColumnExpectations:
    """Test column expectations for Databricks datasources"""

    @parameterize_batch_for_data_sources(
        data_source_configs=[
            DatabricksDatasourceTestConfig(table_name=TEST_TABLE_NAME.lower()),
        ],
        data=pd.DataFrame({"unquoted_lower_col": ["test_value"]}),
    )
    def test_unquoted_lower_col(self, batch_for_datasource):
        """Test Databricks column expectation for unquoted_lower_col"""
        _run_column_expectation_test(batch_for_datasource, "databricks", "unquoted_lower_col")

    @parameterize_batch_for_data_sources(
        data_source_configs=[
            DatabricksDatasourceTestConfig(table_name=TEST_TABLE_NAME.lower()),
        ],
        data=pd.DataFrame({"UNQUOTED_UPPER_COL": ["test_value"]}),
    )
    def test_unquoted_upper_col(self, batch_for_datasource):
        """Test Databricks column expectation for unquoted_upper_col"""
        _run_column_expectation_test(batch_for_datasource, "databricks", "UNQUOTED_UPPER_COL")

    @parameterize_batch_for_data_sources(
        data_source_configs=[
            DatabricksDatasourceTestConfig(table_name=TEST_TABLE_NAME.lower()),
        ],
        data=pd.DataFrame({"quoted_lower_col": ["test_value"]}),
    )
    def test_quoted_lower_col(self, batch_for_datasource):
        """Test Databricks column expectation for quoted_lower_col"""
        _run_column_expectation_test(batch_for_datasource, "databricks", "quoted_lower_col")

    @parameterize_batch_for_data_sources(
        data_source_configs=[
            DatabricksDatasourceTestConfig(table_name=TEST_TABLE_NAME.lower()),
        ],
        data=pd.DataFrame({"QUOTED_UPPER_COL": ["test_value"]}),
    )
    def test_quoted_upper_col(self, batch_for_datasource):
        """Test Databricks column expectation for quoted_upper_col"""
        _run_column_expectation_test(batch_for_datasource, "databricks", "QUOTED_UPPER_COL")

    @parameterize_batch_for_data_sources(
        data_source_configs=[
            DatabricksDatasourceTestConfig(table_name=TEST_TABLE_NAME.lower()),
        ],
        data=pd.DataFrame({"quotedMixed": ["test_value"]}),
    )
    def test_quotedmixed(self, batch_for_datasource):
        """Test Databricks column expectation for quotedmixed"""
        _run_column_expectation_test(batch_for_datasource, "databricks", "quotedMixed")

    @parameterize_batch_for_data_sources(
        data_source_configs=[
            DatabricksDatasourceTestConfig(table_name=TEST_TABLE_NAME.lower()),
        ],
        data=pd.DataFrame({"quoted.w.dots": ["test_value"]}),
    )
    def test_quoted_w_dots(self, batch_for_datasource):
        """Test Databricks column expectation for quoted.w.dots"""
        _run_column_expectation_test(batch_for_datasource, "databricks", "quoted.w.dots")

    @parameterize_batch_for_data_sources(
        data_source_configs=[
            DatabricksDatasourceTestConfig(table_name=TEST_TABLE_NAME.lower()),
        ],
        data=pd.DataFrame({"QUOTED.W.DOTS": ["test_value"]}),
    )
    def test_quoted_w_dots_upper(self, batch_for_datasource):
        """Test Databricks column expectation for QUOTED.W.DOTS"""
        _run_column_expectation_test(batch_for_datasource, "databricks", "QUOTED.W.DOTS")



from __future__ import annotations

import copy
import datetime
import logging
import os
import warnings
from functools import reduce
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
    cast,
    overload,
)

from dateutil.parser import parse

from great_expectations._docs_decorators import deprecated_argument
from great_expectations.compatibility import py4j, pyspark
from great_expectations.compatibility.pyspark import (
    functions as F,
)
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.batch import BatchMarkers
from great_expectations.core.batch_spec import (
    AzureBatchSpec,
    BatchSpec,
    GlueDataCatalogBatchSpec,
    PathBatchSpec,
    RuntimeDataBatchSpec,
)
from great_expectations.core.id_dict import IDDict, IDDictID
from great_expectations.core.metric_domain_types import (
    MetricDomainTypes,  # noqa: TC001 # FIXME CoP
)
from great_expectations.core.util import AzureUrl
from great_expectations.exceptions import (
    BatchSpecError,
    ExecutionEngineError,
    GreatExpectationsError,
    ValidationError,
)
from great_expectations.exceptions import exceptions as gx_exceptions
from great_expectations.execution_engine import ExecutionEngine
from great_expectations.execution_engine.execution_engine import (
    MetricComputationConfiguration,  # noqa: TC001 # FIXME CoP
    PartitionDomainKwargs,  # noqa: TC001 # FIXME CoP
)
from great_expectations.execution_engine.partition_and_sample.sparkdf_data_partitioner import (
    SparkDataPartitioner,
)
from great_expectations.execution_engine.partition_and_sample.sparkdf_data_sampler import (
    SparkDataSampler,
)
from great_expectations.execution_engine.sparkdf_batch_data import SparkDFBatchData
from great_expectations.expectations.model_field_types import (
    CONDITION_PARSER_GREAT_EXPECTATIONS,
    CONDITION_PARSER_GREAT_EXPECTATIONS_DEPRECATED,
    CONDITION_PARSER_SPARK,
)
from great_expectations.expectations.row_conditions import (
    RowCondition,
    RowConditionParserType,
    parse_condition_to_spark,
)
from great_expectations.util import convert_to_json_serializable  # noqa: TID251 # FIXME CoP
from great_expectations.validator.computed_metric import MetricValue  # noqa: TC001 # FIXME CoP

if TYPE_CHECKING:
    from great_expectations.datasource.fluent.spark_datasource import SparkConfig
    from great_expectations.validator.metric_configuration import (
        MetricConfiguration,
        MetricConfigurationID,
    )

logger = logging.getLogger(__name__)


def apply_dateutil_parse(column):
    assert len(column.columns) == 1, "Expected DataFrame with 1 column"
    col_name = column.columns[0]
    _udf = F.udf(parse, pyspark.types.TimestampType())
    return column.withColumn(col_name, _udf(col_name))


@deprecated_argument(
    argument_name="force_reuse_spark_context",
    version="1.0",
    message="The force_reuse_spark_context attribute is no longer part of any Spark Datasource classes. "  # noqa: E501 # FIXME CoP
    "The existing Spark context will be reused if possible. If a spark_config is passed that doesn't match "  # noqa: E501 # FIXME CoP
    "the existing config, the context will be stopped and restarted in local environments only.",
)
class SparkDFExecutionEngine(ExecutionEngine):
    """SparkDFExecutionEngine instantiates the ExecutionEngine API to support computations using Spark platform.

    This class holds an attribute `spark_df` which is a spark.sql.DataFrame.

    Constructor builds a SparkDFExecutionEngine, using provided configuration parameters.

    Args:
        *args: Positional arguments for configuring SparkDFExecutionEngine
        persist: If True (default), then creation of the Spark DataFrame is done outside this class
        spark_config: Dictionary of Spark configuration options. If there is an existing Spark context,
          the spark_config will be used to update that context in environments that allow it. In local
          environments the Spark context will be stopped and restarted with the new spark_config.
        spark: A PySpark Session used to set the SparkDFExecutionEngine being configured. Will override
          spark_config if provided.
        force_reuse_spark_context: If True then utilize existing SparkSession if it exists and is active
        **kwargs: Keyword arguments for configuring SparkDFExecutionEngine

    For example:
    ```python
        name: str = "great_expectations-ee-config"
        spark_config: Dict[str, str] = {
        "spark.app.name": name,
        "spark.sql.catalogImplementation": "hive",
        "spark.executor.memory": "512m",
        }
        execution_engine = SparkDFExecutionEngine(spark_config=spark_config)
        spark_session: SparkSession = execution_engine.spark
    ```

    --ge-feature-maturity-info--

        id: validation_engine_pyspark_self_managed
        title: Validation Engine - pyspark - Self-Managed
        icon:
        short_description: Use Spark DataFrame to validate data
        description: Use Spark DataFrame to validate data
        how_to_guide_url: https://docs.greatexpectations.io/en/latest/how_to_guides/creating_batches/how_to_load_a_spark_dataframe_as_a_batch.html
        maturity: Production
        maturity_details:
            api_stability: Stable
            implementation_completeness: Moderate
            unit_test_coverage: Complete
            integration_infrastructure_test_coverage: N/A -> see relevant Datasource evaluation
            documentation_completeness: Complete
            bug_risk: Low/Moderate
            expectation_completeness: Moderate

        id: validation_engine_databricks
        title: Validation Engine - Databricks
        icon:
        short_description: Use Spark DataFrame in a Databricks cluster to validate data
        description: Use Spark DataFrame in a Databricks cluster to validate data
        how_to_guide_url: https://docs.greatexpectations.io/en/latest/how_to_guides/creating_batches/how_to_load_a_spark_dataframe_as_a_batch.html
        maturity: Beta
        maturity_details:
            api_stability: Stable
            implementation_completeness: Low (dbfs-specific handling)
            unit_test_coverage: N/A -> implementation not different
            integration_infrastructure_test_coverage: Minimal (we've tested a bit, know others have used it)
            documentation_completeness: Moderate (need docs on managing project configuration via dbfs/etc.)
            bug_risk: Low/Moderate
            expectation_completeness: Moderate

        id: validation_engine_emr_spark
        title: Validation Engine - EMR - Spark
        icon:
        short_description: Use Spark DataFrame in an EMR cluster to validate data
        description: Use Spark DataFrame in an EMR cluster to validate data
        how_to_guide_url: https://docs.greatexpectations.io/en/latest/how_to_guides/creating_batches/how_to_load_a_spark_dataframe_as_a_batch.html
        maturity: Experimental
        maturity_details:
            api_stability: Stable
            implementation_completeness: Low (need to provide guidance on "known good" paths, and we know there are many "knobs" to tune that we have not explored/tested)
            unit_test_coverage: N/A -> implementation not different
            integration_infrastructure_test_coverage: Unknown
            documentation_completeness: Low (must install specific/latest version but do not have docs to that effect or of known useful paths)
            bug_risk: Low/Moderate
            expectation_completeness: Moderate

        id: validation_engine_spark_other
        title: Validation Engine - Spark - Other
        icon:
        short_description: Use Spark DataFrame to validate data
        description: Use Spark DataFrame to validate data
        how_to_guide_url: https://docs.greatexpectations.io/en/latest/how_to_guides/creating_batches/how_to_load_a_spark_dataframe_as_a_batch.html
        maturity: Experimental
        maturity_details:
            api_stability: Stable
            implementation_completeness: Other (we haven't tested possibility, known glue deployment)
            unit_test_coverage: N/A -> implementation not different
            integration_infrastructure_test_coverage: Unknown
            documentation_completeness: Low (must install specific/latest version but do not have docs to that effect or of known useful paths)
            bug_risk: Low/Moderate
            expectation_completeness: Moderate

    --ge-feature-maturity-info--
    """  # noqa: E501 # FIXME CoP

    recognized_batch_definition_keys = {"limit"}

    recognized_batch_spec_defaults = {
        "reader_method",
        "reader_options",
    }

    def __init__(
        self,
        *args,
        persist: bool = True,
        spark_config: Optional[dict] = None,
        spark: Optional[pyspark.SparkSession] = None,
        force_reuse_spark_context: Optional[bool] = None,
        **kwargs,
    ) -> None:
        self._persist = persist

        spark_config = spark_config or {}
        self.spark: pyspark.SparkSession
        if spark:
            self.spark = spark
        else:
            self.spark = SparkDFExecutionEngine.get_or_create_spark_session(
                spark_config=spark_config,
            )

        azure_options: dict = kwargs.pop("azure_options", {})
        self._azure_options = azure_options

        if force_reuse_spark_context is not None:
            # deprecated-v1.0.0
            warnings.warn(
                "force_reuse_spark_context is deprecated and will be removed in version 1.0. "
                "In environments that allow it, the existing Spark context will be reused, adding the "  # noqa: E501 # FIXME CoP
                "spark_config options that have been passed. If the Spark context cannot be updated with "  # noqa: E501 # FIXME CoP
                "the spark_config, the context will be stopped and restarted with the new spark_config.",  # noqa: E501 # FIXME CoP
                category=DeprecationWarning,
            )
        super().__init__(*args, **kwargs)

        self._config.update(
            {
                "persist": self._persist,
                "spark_config": spark_config,
                "azure_options": azure_options,
            }
        )

        self._data_partitioner = SparkDataPartitioner()
        self._data_sampler = SparkDataSampler()

    @property
    def dataframe(self) -> pyspark.DataFrame:
        """If a batch has been loaded, returns a Spark Dataframe containing the data within the loaded batch"""  # noqa: E501 # FIXME CoP
        if self.batch_manager.active_batch_data is None:
            raise ValueError("Batch has not been loaded - please run load_batch() to load a batch.")  # noqa: TRY003 # FIXME CoP

        return cast("SparkDFBatchData", self.batch_manager.active_batch_data).dataframe

    @staticmethod
    def get_or_create_spark_session(
        spark_config: Optional[SparkConfig] = None,
    ) -> pyspark.SparkSession:
        """Obtains Spark session if it already exists; otherwise creates Spark session and returns it to caller.

        Args:
            spark_config: Dictionary containing Spark configuration (string-valued keys mapped to string-valued properties).

        Returns:
            SparkSession
        """  # noqa: E501 # FIXME CoP
        spark_config = spark_config or {}

        spark_session: pyspark.SparkSession
        try:
            spark_session = pyspark.SparkConnectSession.builder.getOrCreate()
        except (ModuleNotFoundError, ValueError, KeyError):
            spark_session = pyspark.SparkSession.builder.getOrCreate()

        return SparkDFExecutionEngine._get_session_with_spark_config(
            spark_config=spark_config,
            spark_session=spark_session,
        )

    @staticmethod
    def _get_session_with_spark_config(
        spark_session: pyspark.SparkSession,
        spark_config: dict,
    ) -> pyspark.SparkSession:
        """Attempts to apply spark_config to a SparkSession by either:
             1. Updating the existing SparkSession with spark_config values
             2. Restarting the existing SparkSession and applying only spark_config

          If a spark_config option is unable to be set, a warning is raised.

        Args:
            spark_session: An existing pyspark.SparkSession.
            spark_config: A dictionary of SparkSession.Builder.config objects.

        Returns:
            SparkSession
        """
        stopped: bool
        (
            spark_session,
            stopped,
        ) = SparkDFExecutionEngine._try_update_or_stop_misconfigured_spark_session(
            spark_session=spark_session,
            spark_config=spark_config,
        )

        if stopped:
            spark_session = SparkDFExecutionEngine._start_spark_session_with_spark_config(
                spark_session=spark_session,
                spark_config=spark_config,
            )

        return spark_session

    @staticmethod
    def _start_spark_session_with_spark_config(
        spark_session: pyspark.SparkSession,
        spark_config: dict,
    ) -> pyspark.SparkSession:
        builder = spark_session.builder
        for key, value in spark_config.items():
            if key == "spark.app.name":
                builder.appName(value)
            else:
                builder.config(key, value)

        return builder.getOrCreate()

    @staticmethod
    def _session_is_not_stoppable(
        spark_session: pyspark.SparkSession,
    ) -> bool:
        return (
            pyspark.SparkConnectSession  # type: ignore[truthy-function]  # returns false if module is not installed
            and isinstance(spark_session, pyspark.SparkConnectSession)
        ) or (
            os.environ.get("DATABRICKS_RUNTIME_VERSION") is not None  # noqa: TID251 # FIXME CoP
        )

    @staticmethod
    def _try_update_or_stop_misconfigured_spark_session(
        spark_session: pyspark.SparkSession,
        spark_config: dict,
    ) -> tuple[pyspark.SparkSession, bool]:
        """Tries to update the SparkSession if it doesn't have the options specified in spark_config set.
        If updates fail, and the SparkSession can be stopped, it will be stopped.
        If the SparkSession cannot be stopped, it will be returned unaltered.

        Warns if the SparkSession was stopped or a config option could not be set.

        Returns:
            SparkSession, Boolean specifying if SparkSession is stopped
        """  # noqa: E501 # FIXME CoP
        stopped = False
        warning_messages = []
        for key, value in spark_config.items():
            # if the user set a spark_config option that doesn't match the existing session
            # try to update it, otherwise stop the spark session
            try:
                # conf.get will look first at the runtime conf and then at the sparkContext conf
                try:
                    current_value = spark_session.conf.get(key)
                # Py4J Java Error can be raised if the option has not been set on the context at all
                except py4j.protocol.Py4JJavaError:
                    current_value = None
                if key != "spark.app.name" and (current_value != value or current_value is None):
                    # attempts to update the runtime config
                    spark_session.conf.set(key, value)
                elif key == "spark.app.name" and spark_session.sparkContext.appName != value:
                    spark_session.sparkContext.appName = value
            # attribute error can be raised for connect sessions that haven't implemented a conf for sparkContext method  # noqa: E501 # FIXME CoP
            # analysis exception can be raised in environments that don't allow updating config of that option  # noqa: E501 # FIXME CoP
            except (
                pyspark.PySparkAttributeError,
                pyspark.AnalysisException,
            ):
                if SparkDFExecutionEngine._session_is_not_stoppable(spark_session=spark_session):
                    warning_messages.append(
                        f"Passing spark_config option `{key}` had no effect, because in this environment "  # noqa: E501 # FIXME CoP
                        "it is not modifiable and the Spark Session cannot be restarted."
                    )
                else:
                    spark_session.stop()
                    stopped = True
                    warning_messages.append(
                        f"Spark Session was restarted, because `{key}` "
                        "is not modifiable in this environment."
                    )
                    break

        for message in warning_messages:
            warnings.warn(
                message=message,
                category=RuntimeWarning,
            )

        return spark_session, stopped

    @override
    def load_batch_data(  # type: ignore[override] # FIXME CoP
        self, batch_id: str, batch_data: Union[SparkDFBatchData, pyspark.DataFrame]
    ) -> None:
        if pyspark.DataFrame and isinstance(batch_data, pyspark.DataFrame):  # type: ignore[truthy-function] # FIXME CoP
            batch_data = SparkDFBatchData(self, batch_data)
        elif not isinstance(batch_data, SparkDFBatchData):
            raise GreatExpectationsError(  # noqa: TRY003 # FIXME CoP
                "SparkDFExecutionEngine requires batch data that is either a DataFrame or a SparkDFBatchData object"  # noqa: E501 # FIXME CoP
            )

        if self._persist:
            batch_data.dataframe.persist()

        super().load_batch_data(batch_id=batch_id, batch_data=batch_data)

    @override
    def get_batch_data_and_markers(  # noqa: C901, PLR0912, PLR0915 # FIXME CoP
        self, batch_spec: BatchSpec
    ) -> Tuple[Any, BatchMarkers]:  # batch_data
        # We need to build a batch_markers to be used in the dataframe
        batch_markers = BatchMarkers(
            {
                "ge_load_time": datetime.datetime.now(datetime.timezone.utc).strftime(
                    "%Y%m%dT%H%M%S.%fZ"
                )
            }
        )

        """
        As documented in Azure DataConnector implementations, Pandas and Spark execution engines utilize separate path
        formats for accessing Azure Blob Storage service.  However, Pandas and Spark execution engines utilize identical
        path formats for accessing all other supported cloud storage services (AWS S3 and Google Cloud Storage).
        Moreover, these formats (encapsulated in S3BatchSpec and GCSBatchSpec) extend PathBatchSpec (common to them).
        Therefore, at the present time, all cases with the exception of Azure Blob Storage, are handled generically.
        """  # noqa: E501 # FIXME CoP

        batch_data: Any
        reader_method: str
        reader_options: dict
        path: str
        schema: Optional[Union[pyspark.types.StructType, dict, str]]
        reader: pyspark.DataFrameReader
        reader_fn: Callable
        if isinstance(batch_spec, RuntimeDataBatchSpec):
            # batch_data != None is already checked when RuntimeDataBatchSpec is instantiated
            batch_data = batch_spec.batch_data
            if isinstance(batch_data, str):
                raise gx_exceptions.ExecutionEngineError(  # noqa: TRY003 # FIXME CoP
                    f"""SparkDFExecutionEngine has been passed a string type batch_data, "{batch_data}", which is \
illegal.  Please check your config."""  # noqa: E501 # FIXME CoP
                )
            batch_spec.batch_data = "SparkDataFrame"

        elif isinstance(batch_spec, AzureBatchSpec):
            reader_method = batch_spec.reader_method
            reader_options = batch_spec.reader_options or {}
            path = batch_spec.path
            azure_url = AzureUrl(path)
            # TODO <WILL> 202209 - Add `schema` definition to Azure like PathBatchSpec below (GREAT-1224)  # noqa: E501 # FIXME CoP
            try:
                credential = self._azure_options.get("credential")
                storage_account_url = azure_url.account_url
                if credential:
                    self.spark.conf.set(
                        "fs.wasb.impl",
                        "org.apache.hadoop.fs.azure.NativeAzureFileSystem",
                    )
                    self.spark.conf.set(f"fs.azure.account.key.{storage_account_url}", credential)
                reader = self.spark.read.options(**reader_options)
                reader_fn = self._get_reader_fn(
                    reader=reader,
                    reader_method=reader_method,
                    path=path,
                )
                batch_data = reader_fn(path)
            except AttributeError:
                raise ExecutionEngineError(  # noqa: TRY003 # FIXME CoP
                    """
                    Unable to load pyspark. Pyspark is required for SparkDFExecutionEngine.
                    """
                )

        elif isinstance(batch_spec, (PathBatchSpec, GlueDataCatalogBatchSpec)):
            reader_method = batch_spec.reader_method
            reader_options = batch_spec.reader_options or {}
            path = batch_spec.path
            schema = reader_options.get("schema")

            # schema can be a dict if it has been through serialization step,
            # either as part of the datasource configuration, or checkpoint config
            if isinstance(schema, dict):
                schema = pyspark.types.StructType.fromJson(schema)

            # this can happen if we have not converted schema into json at Datasource-config level
            elif isinstance(schema, str):
                raise gx_exceptions.ExecutionEngineError(  # noqa: TRY003 # FIXME CoP
                    """
                    Spark schema was not properly serialized.
                    Please run the .jsonValue() method on the schema object before loading into GX.
                    schema: your_schema.jsonValue()
                    """
                )
            # noinspection PyUnresolvedReferences
            try:
                if schema:
                    reader = self.spark.read.schema(schema).options(**reader_options)
                else:
                    reader = self.spark.read.options(**reader_options)

                reader_fn = self._get_reader_fn(
                    reader=reader,
                    reader_method=reader_method,
                    path=path,
                )
                batch_data = reader_fn(path)
            except AttributeError:
                raise ExecutionEngineError(  # noqa: TRY003 # FIXME CoP
                    """
                    Unable to load pyspark. Pyspark is required for SparkDFExecutionEngine.
                    """
                )
            # pyspark will raise an AnalysisException error if path is incorrect
            except pyspark.AnalysisException:
                raise ExecutionEngineError(  # noqa: TRY003 # FIXME CoP
                    f"""Unable to read in batch from the following path: {path}. Please check your configuration."""  # noqa: E501 # FIXME CoP
                )

        else:
            raise BatchSpecError(  # noqa: TRY003 # FIXME CoP
                """
                Invalid batch_spec: batch_data is required for a SparkDFExecutionEngine to operate.
                """
            )

        batch_data = self._apply_partitioning_and_sampling_methods(batch_spec, batch_data)
        typed_batch_data = SparkDFBatchData(execution_engine=self, dataframe=batch_data)

        return typed_batch_data, batch_markers

    def _apply_partitioning_and_sampling_methods(self, batch_spec, batch_data):
        # Note this is to get a batch from tables in AWS Glue Data Catalog by its partitions
        partitions: Optional[List[str]] = batch_spec.get("partitions")
        if partitions:
            batch_data = self._data_partitioner.partition_on_multi_column_values(
                df=batch_data,
                column_names=partitions,
                batch_identifiers=batch_spec.get("batch_identifiers"),
            )

        partitioner_method_name: Optional[str] = batch_spec.get("partitioner_method")
        if partitioner_method_name:
            partitioner_fn: Callable = self._data_partitioner.get_partitioner_method(
                partitioner_method_name
            )
            partitioner_kwargs: dict = batch_spec.get("partitioner_kwargs") or {}
            batch_data = partitioner_fn(batch_data, **partitioner_kwargs)

        sampler_method_name: Optional[str] = batch_spec.get("sampling_method")
        if sampler_method_name:
            sampling_fn: Callable = self._data_sampler.get_sampler_method(sampler_method_name)
            batch_data = sampling_fn(batch_data, batch_spec)

        return batch_data

    # TODO: <Alex>Similar to Abe's note in PandasExecutionEngine: Any reason this shouldn't be a private method?</Alex>  # noqa: E501 # FIXME CoP
    @staticmethod
    def guess_reader_method_from_path(path: str):
        """
        Based on a given filepath, decides a reader method. Currently supports tsv, csv, and parquet. If none of these
        file extensions are used, returns ExecutionEngineError stating that it is unable to determine the current path.

        Args:
            path - A given file path

        Returns:
            A dictionary entry of format {'reader_method': reader_method}

        """  # noqa: E501 # FIXME CoP
        path = path.lower()
        if path.endswith(".csv") or path.endswith(".tsv"):
            return "csv"
        elif path.endswith(".parquet") or path.endswith(".parq") or path.endswith(".pqt"):
            return "parquet"

        raise ExecutionEngineError(f"Unable to determine reader method from path: {path}")  # noqa: TRY003 # FIXME CoP

    @overload
    def _get_reader_fn(
        self, reader, reader_method: str = ..., path: Optional[str] = ...
    ) -> Callable: ...

    @overload
    def _get_reader_fn(self, reader, reader_method: None = ..., path: str = ...) -> Callable: ...

    def _get_reader_fn(self, reader, reader_method=None, path=None) -> Callable:
        """Static helper for providing reader_fn

        Args:
            reader: the base spark reader to use; this should have had reader_options applied already
            reader_method: the name of the reader_method to use, if specified
            path (str): the path to use to guess reader_method if it was not specified

        Returns:
            ReaderMethod to use for the filepath

        """  # noqa: E501 # FIXME CoP
        if reader_method is None and path is None:
            raise ExecutionEngineError(  # noqa: TRY003 # FIXME CoP
                "Unable to determine spark reader function without reader_method or path"
            )

        if reader_method is None:
            reader_method = self.guess_reader_method_from_path(path=path)

        reader_method_op: str = reader_method.lower()
        try:
            if reader_method_op == "delta":
                return reader.format(reader_method_op).load
            return getattr(reader, reader_method_op)
        except AttributeError:
            raise ExecutionEngineError(  # noqa: TRY003 # FIXME CoP
                f"Unable to find reader_method {reader_method} in spark.",
            )

    @override
    def get_domain_records(  # noqa: C901, PLR0912, PLR0915 # FIXME CoP
        self,
        domain_kwargs: dict,
    ) -> "pyspark.DataFrame":  # noqa F821
        """Uses the given Domain kwargs (which include row_condition, condition_parser, and ignore_row_if directives) to obtain and/or query a batch.

        Args:
            domain_kwargs (dict) - A dictionary consisting of the Domain kwargs specifying which data to obtain

        Returns:
            A DataFrame (the data on which to compute returned in the format of a Spark DataFrame)
        """  # noqa: E501 # FIXME CoP
        """
        # TODO: <Alex>Docusaurus run fails, unless "pyspark.DataFrame" type hint above is enclosed in quotes.
        This may be caused by it becoming great_expectations.compatibility.not_imported.NotImported when pyspark is not installed.
        </Alex>
        """  # noqa: E501 # FIXME CoP
        table = domain_kwargs.get("table", None)
        if table:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                "SparkDFExecutionEngine does not currently support multiple named tables."
            )

        batch_id = domain_kwargs.get("batch_id")
        if batch_id is None:
            # We allow no batch id specified if there is only one batch
            if self.batch_manager.active_batch_data:
                data = cast("SparkDFBatchData", self.batch_manager.active_batch_data).dataframe
            else:
                raise ValidationError(  # noqa: TRY003 # FIXME CoP
                    "No batch is specified, but could not identify a loaded batch."
                )
        else:  # noqa: PLR5501 # FIXME CoP
            if batch_id in self.batch_manager.batch_data_cache:
                data = cast(
                    "SparkDFBatchData", self.batch_manager.batch_data_cache[batch_id]
                ).dataframe
            else:
                raise ValidationError(f"Unable to find batch with batch_id {batch_id}")  # noqa: TRY003 # FIXME CoP

        # Filtering by row condition.
        row_condition = domain_kwargs.get("row_condition", None)
        if row_condition:
            condition_parser = domain_kwargs.get("condition_parser", None)
            if condition_parser == CONDITION_PARSER_SPARK:
                data = data.filter(row_condition)
            elif condition_parser in [
                CONDITION_PARSER_GREAT_EXPECTATIONS,
                CONDITION_PARSER_GREAT_EXPECTATIONS_DEPRECATED,
            ]:
                parsed_condition = parse_condition_to_spark(row_condition)
                data = data.filter(parsed_condition)
            else:
                raise GreatExpectationsError(  # noqa: TRY003 # FIXME CoP
                    f"unrecognized condition_parser {condition_parser!s} for Spark execution engine"
                )

        # Filtering by filter_conditions
        filter_conditions: List[RowCondition] = domain_kwargs.get("filter_conditions", [])
        if len(filter_conditions) > 0:
            filter_condition = self._combine_row_conditions(filter_conditions)
            data = data.filter(filter_condition.condition)

        if "column" in domain_kwargs:
            return data

        # Filtering by ignore_row_if directive
        if (
            "column_A" in domain_kwargs
            and "column_B" in domain_kwargs
            and "ignore_row_if" in domain_kwargs
        ):
            # noinspection PyPep8Naming
            column_A_name = domain_kwargs["column_A"]
            # noinspection PyPep8Naming
            column_B_name = domain_kwargs["column_B"]

            ignore_row_if = domain_kwargs["ignore_row_if"]
            if ignore_row_if == "both_values_are_missing":
                ignore_condition = F.col(column_A_name).isNull() & F.col(column_B_name).isNull()
                data = data.filter(~ignore_condition)
            elif ignore_row_if == "either_value_is_missing":
                ignore_condition = F.col(column_A_name).isNull() | F.col(column_B_name).isNull()
                data = data.filter(~ignore_condition)
            else:  # noqa: PLR5501 # FIXME CoP
                if ignore_row_if != "neither":
                    raise ValueError(f'Unrecognized value of ignore_row_if ("{ignore_row_if}").')  # noqa: TRY003 # FIXME CoP

            return data

        if "column_list" in domain_kwargs and "ignore_row_if" in domain_kwargs:
            column_list = domain_kwargs["column_list"]
            ignore_row_if = domain_kwargs["ignore_row_if"]
            if ignore_row_if == "all_values_are_missing":
                conditions = [F.col(column_name).isNull() for column_name in column_list]
                ignore_condition = reduce(lambda a, b: a & b, conditions)
                data = data.filter(~ignore_condition)
            elif ignore_row_if == "any_value_is_missing":
                conditions = [F.col(column_name).isNull() for column_name in column_list]
                ignore_condition = reduce(lambda a, b: a | b, conditions)
                data = data.filter(~ignore_condition)
            else:  # noqa: PLR5501 # FIXME CoP
                if ignore_row_if != "never":
                    raise ValueError(f'Unrecognized value of ignore_row_if ("{ignore_row_if}").')  # noqa: TRY003 # FIXME CoP

            return data

        return data

    @staticmethod
    def _combine_row_conditions(row_conditions: List[RowCondition]) -> RowCondition:
        """Combine row conditions using AND if condition_type is SPARK_SQL

        Note, although this method does not currently use `self` internally we
        are not marking as @staticmethod since it is meant to only be called
        internally in this class.

        Args:
            row_conditions: Row conditions of type Spark

        Returns:
            Single Row Condition combined
        """
        assert all(
            condition.condition_type == RowConditionParserType.SPARK_SQL
            for condition in row_conditions
        ), "All row conditions must have type SPARK_SQL"
        conditions: List[str] = [row_condition.condition for row_condition in row_conditions]
        joined_condition: str = " AND ".join(conditions)
        return RowCondition(
            condition=joined_condition, condition_type=RowConditionParserType.SPARK_SQL
        )

    @override
    def get_compute_domain(
        self,
        domain_kwargs: dict,
        domain_type: Union[str, MetricDomainTypes],
        accessor_keys: Optional[Iterable[str]] = None,
    ) -> Tuple["pyspark.DataFrame", dict, dict]:  # noqa F821
        """Uses a DataFrame and Domain kwargs (which include a row condition and a condition parser) to obtain and/or query a Batch of data.

        Returns in the format of a Spark DataFrame along with Domain arguments required for computing.  If the Domain \
        is a single column, this is added to 'accessor Domain kwargs' and used for later access.

        Args:
            domain_kwargs (dict): a dictionary consisting of the Domain kwargs specifying which data to obtain
            domain_type (str or MetricDomainTypes): an Enum value indicating which metric Domain the user would like \
            to be using, or a corresponding string value representing it.  String types include "identity", "column", \
            "column_pair", "table" and "other".  Enum types include capitalized versions of these from the class \
            MetricDomainTypes.
            accessor_keys (str iterable): keys that are part of the compute Domain but should be ignored when \
            describing the Domain and simply transferred with their associated values into accessor_domain_kwargs.

        Returns:
            A tuple including:
              - a DataFrame (the data on which to compute)
              - a dictionary of compute_domain_kwargs, describing the DataFrame
              - a dictionary of accessor_domain_kwargs, describing any accessors needed to
                identify the Domain within the compute domain
        """  # noqa: E501 # FIXME CoP
        """
        # TODO: <Alex>Docusaurus run fails, unless "pyspark.DataFrame" type hint above is enclosed in quotes.
        This may be caused by it becoming great_expectations.compatibility.not_imported.NotImported when pyspark is not installed.
        </Alex>
        """  # noqa: E501 # FIXME CoP
        table: str = domain_kwargs.get("table", None)
        if table:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                "SparkDFExecutionEngine does not currently support multiple named tables."
            )

        data: pyspark.DataFrame = self.get_domain_records(domain_kwargs=domain_kwargs)

        partitioned_domain_kwargs: PartitionDomainKwargs = self._partition_domain_kwargs(
            domain_kwargs, domain_type, accessor_keys
        )

        return (
            data,
            partitioned_domain_kwargs.compute,
            partitioned_domain_kwargs.accessor,
        )

    def add_column_row_condition(  # type: ignore[explicit-override] # FIXME
        self, domain_kwargs, column_name=None, filter_null=True, filter_nan=False
    ):
        # We explicitly handle filter_nan & filter_null for spark using a spark-native condition

        new_domain_kwargs = copy.deepcopy(domain_kwargs)
        assert "column" in domain_kwargs or column_name is not None
        if column_name is not None:
            column = column_name
        else:
            column = domain_kwargs["column"]

        filter_conditions: List[RowCondition] = []
        if filter_null:
            filter_conditions.append(
                RowCondition(
                    condition=f"{column} IS NOT NULL",
                    condition_type=RowConditionParserType.SPARK_SQL,
                )
            )
        if filter_nan:
            filter_conditions.append(
                RowCondition(
                    condition=f"NOT isnan({column})",
                    condition_type=RowConditionParserType.SPARK_SQL,
                )
            )

        if not (filter_null or filter_nan):
            logger.warning(
                "add_column_row_condition called without specifying a desired row condition"
            )

        new_domain_kwargs.setdefault("filter_conditions", []).extend(filter_conditions)

        return new_domain_kwargs

    @override
    def resolve_metric_bundle(
        self,
        metric_fn_bundle: Iterable[MetricComputationConfiguration],
    ) -> dict[MetricConfigurationID, MetricValue]:
        """For every metric in a set of Metrics to resolve, obtains necessary metric keyword arguments and builds
        bundles of the metrics into one large query dictionary so that they are all executed simultaneously. Will fail
        if bundling the metrics together is not possible.

            Args:
                metric_fn_bundle (Iterable[MetricComputationConfiguration]): \
                    "MetricComputationConfiguration" contains MetricProvider's MetricConfiguration (its unique identifier),
                    its metric provider function (the function that actually executes the metric), and arguments to pass
                    to metric provider function (dictionary of metrics defined in registry and corresponding arguments).

            Returns:
                A dictionary of "MetricConfiguration" IDs and their corresponding fully resolved values for domains.
        """  # noqa: E501 # FIXME CoP
        resolved_metrics: dict[MetricConfigurationID, MetricValue] = {}

        res: List[pyspark.Row]

        aggregates: dict[IDDictID, dict] = {}

        aggregate: dict

        domain_id: IDDictID

        bundled_metric_configuration: MetricComputationConfiguration
        for bundled_metric_configuration in metric_fn_bundle:
            metric_to_resolve: MetricConfiguration = (
                bundled_metric_configuration.metric_configuration
            )
            metric_fn: Any = bundled_metric_configuration.metric_fn
            compute_domain_kwargs: dict = bundled_metric_configuration.compute_domain_kwargs or {}
            if not isinstance(compute_domain_kwargs, IDDict):
                compute_domain_kwargs = IDDict(compute_domain_kwargs)

            domain_id = compute_domain_kwargs.to_id()
            if domain_id not in aggregates:
                aggregates[domain_id] = {
                    "column_aggregates": [],
                    "metric_ids": [],
                    "domain_kwargs": compute_domain_kwargs,
                }

            aggregates[domain_id]["column_aggregates"].append(metric_fn)
            aggregates[domain_id]["metric_ids"].append(metric_to_resolve.id)

        for aggregate in aggregates.values():
            domain_kwargs: dict = aggregate["domain_kwargs"]
            df: pyspark.DataFrame = self.get_domain_records(domain_kwargs=domain_kwargs)

            assert len(aggregate["column_aggregates"]) == len(aggregate["metric_ids"])

            res = df.agg(*aggregate["column_aggregates"]).collect()

            logger.debug(
                f"SparkDFExecutionEngine computed {len(res[0])} metrics on domain_id {IDDict(domain_kwargs).to_id()}"  # noqa: E501 # FIXME CoP
            )

            assert len(res) == 1, "all bundle-computed metrics must be single-value statistics"
            assert len(aggregate["metric_ids"]) == len(res[0]), (
                "unexpected number of metrics returned"
            )

            idx: int
            metric_id: MetricConfigurationID
            for idx, metric_id in enumerate(aggregate["metric_ids"]):
                # Converting DataFrame.collect() results into JSON-serializable format produces simple data types,  # noqa: E501 # FIXME CoP
                # amenable for subsequent post-processing by higher-level "Metric" and "Expectation" layers.  # noqa: E501 # FIXME CoP
                resolved_metrics[metric_id] = convert_to_json_serializable(data=res[0][idx])

        return resolved_metrics

    def head(self, n=5):
        """Returns dataframe head. Default is 5"""
        return self.dataframe.limit(n).toPandas()



from __future__ import annotations

import copy
import dataclasses
import functools
import logging
import uuid
import warnings
from abc import ABC, abstractmethod
from pprint import pformat as pf
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Callable,
    ClassVar,
    Dict,
    Final,
    Generic,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Protocol,
    Sequence,
    Set,
    Type,
    TypeVar,
    Union,
    overload,
)

from great_expectations._docs_decorators import public_api
from great_expectations.compatibility import pydantic
from great_expectations.compatibility.pydantic import (
    Field,
    StrictBool,
    StrictInt,
    validate_arguments,
)
from great_expectations.compatibility.pydantic import dataclasses as pydantic_dc
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.batch_definition import BatchDefinition, PartitionerT
from great_expectations.core.config_substitutor import _ConfigurationSubstitutor
from great_expectations.core.result_format import DEFAULT_RESULT_FORMAT
from great_expectations.datasource.fluent.constants import (
    _ASSETS_KEY,
)
from great_expectations.datasource.fluent.fluent_base_model import (
    FluentBaseModel,
    GenericBaseModel,
)
from great_expectations.datasource.fluent.metadatasource import MetaDatasource
from great_expectations.exceptions.exceptions import (
    DataAssetInitializationError,
    DataContextError,
    MissingDataContextError,
)
from great_expectations.metrics.metric import MetaMetric, Metric
from great_expectations.metrics.metric_name import MetricNameSuffix
from great_expectations.metrics.metric_results import (
    MetricErrorResult,
    MetricErrorResultValue,
    MetricResult,
)
from great_expectations.validator.metrics_calculator import (
    MetricsCalculator,
)

logger = logging.getLogger(__name__)
from great_expectations.datasource.fluent.data_connector import (
    DataConnector,
)

if TYPE_CHECKING:
    import pandas as pd
    from typing_extensions import TypeAlias, TypeGuard

    from great_expectations.core.result_format import ResultFormatUnion
    from great_expectations.core.suite_parameters import SuiteParameterDict

    MappingIntStrAny = Mapping[Union[int, str], Any]
    AbstractSetIntStr = AbstractSet[Union[int, str]]
    from great_expectations.core import (
        ExpectationSuite,
        ExpectationSuiteValidationResult,
        ExpectationValidationResult,
    )
    from great_expectations.core.batch import (
        BatchData,
        BatchMarkers,
        LegacyBatchDefinition,
    )
    from great_expectations.core.config_provider import _ConfigurationProvider
    from great_expectations.core.id_dict import BatchSpec
    from great_expectations.data_context import (
        AbstractDataContext as GXDataContext,
    )
    from great_expectations.datasource.fluent import (
        BatchParameters,
        BatchRequest,
    )
    from great_expectations.datasource.fluent.data_connector.batch_filter import BatchSlice
    from great_expectations.datasource.fluent.type_lookup import (
        TypeLookup,
    )
    from great_expectations.expectations.expectation import Expectation
    from great_expectations.validator.computed_metric import MetricValue
    from great_expectations.validator.v1_validator import (
        Validator as V1Validator,
    )

_T = TypeVar("_T")
_MetricResultT = TypeVar("_MetricResultT", bound=MetricResult)


class PartitionerSortingProtocol(Protocol):
    """Interface defining the fields a Partitioner must contain for sorting."""

    sort_ascending: bool

    @property
    def param_names(self) -> Union[list[str], tuple[str, ...]]:
        """The parameter names that specify a batch derived from this partitioner

        For example, for PartitionerYearMonth this returns ["year", "month"]. For more
        examples, please see concrete Partitioner* classes.
        """
        ...


class PartitionerProtocol(PartitionerSortingProtocol, Protocol):
    @property
    def columns(self) -> list[str]:
        """The names of the column used to partition the data"""
        ...

    @property
    def method_name(self) -> str:
        """Returns a partitioner method name.

        The possible values of partitioner method names are defined in the enum,
        great_expectations.execution_engine.partition_and_sample.data_partitioner.PartitionerMethod
        """
        ...

    def partitioner_method_kwargs(self) -> Dict[str, Any]:
        """A shim to our execution engine partitioner methods

        We translate any internal Partitioner state and what is passed in from
        a batch_request to the partitioner_kwargs required by our execution engine.

        Look at Partitioner* classes for concrete examples.
        """
        ...

    def batch_parameters_to_batch_spec_kwarg_identifiers(
        self, options: BatchParameters
    ) -> Dict[str, Any]:
        """Translates `options` to the execution engine batch spec kwarg identifiers

        Arguments:
            options: A BatchRequest.options dictionary that specifies ALL the fields necessary
                     to specify a batch with respect to this partitioner.

        Returns:
            A dictionary that can be added to batch_spec_kwargs["batch_identifiers"].
            This has one of 2 forms:
              1. This category has many parameters are derived from 1 column.
                 These only are datetime partitioners and the batch_spec_kwargs["batch_identifiers"]
                 look like:
                   {column_name: {datepart_1: value, datepart_2: value, ...}
                 where datepart_* are strings like "year", "month", "day". The exact
                 fields depend on the partitioner.

              2. This category has only 1 parameter for each column.
                 This is used for all other partitioners and the
                 batch_spec_kwargs["batch_identifiers"]
                 look like:
                   {column_name_1: value, column_name_2: value, ...}
                 where value is the value of the column after being processed by the partitioner.
                 For example, for the PartitionerModInteger where mod = 3,
                 {"passenger_count": 2}, means the raw passenger count value is in the set:
                 {2, 5, 8, ...} = {2*n + 1 | n is a nonnegative integer }
                 This category was only 1 parameter per column.
        """
        ...


class TestConnectionError(ConnectionError):
    """
    Raised if `.test_connection()` fails to connect to the datasource.
    """

    def __init__(
        self,
        message: str = "Attempt to connect to datasource failed",
        *,
        cause: Exception | None = None,
        addendum: str | None = None,
    ):
        """
        Args:
            `message` base of the error message to be provided to the user.
            `cause` is the original exception that caused the error, the repr of which will be added
                to the error message.
            `addendum` is optional additional information that can be added to the error message.
        """
        self.cause = cause  # not guaranteed to be the same as `self.__cause__`
        self.addendum = addendum
        if cause:
            message += f": due to {cause!r}"
        if addendum:
            message += f": {addendum}"
        super().__init__(message)


class GxDatasourceWarning(UserWarning):
    """
    Warning related to usage or configuration of a Datasource that could lead to
    unexpected behavior.
    """


class GxContextWarning(GxDatasourceWarning):
    """
    Warning related to a Datasource with a missing context.
    Usually because the Datasource was created directly rather than using a
    `context.data_sources` factory method.
    """


class GxSerializationWarning(GxDatasourceWarning):
    pass


BatchMetadata: TypeAlias = Dict[str, Any]


@pydantic_dc.dataclass(frozen=True)
class Sorter:
    key: str
    reverse: bool = False


SortersDefinition: TypeAlias = List[Union[Sorter, str, dict]]


def _is_sorter_list(
    sorters: SortersDefinition,
) -> TypeGuard[list[Sorter]]:
    return len(sorters) == 0 or isinstance(sorters[0], Sorter)


def _is_str_sorter_list(sorters: SortersDefinition) -> TypeGuard[list[str]]:
    return len(sorters) > 0 and isinstance(sorters[0], str)


def _sorter_from_list(sorters: SortersDefinition) -> list[Sorter]:
    if _is_sorter_list(sorters):
        return sorters

    # mypy doesn't successfully type-narrow sorters to a list[str] here, so we use
    # another TypeGuard. We could cast instead which may be slightly faster.
    sring_valued_sorter: str
    if _is_str_sorter_list(sorters):
        return [_sorter_from_str(sring_valued_sorter) for sring_valued_sorter in sorters]

    # This should never be reached because of static typing but is necessary because
    # mypy doesn't know of the if conditions must evaluate to True.
    raise ValueError(  # noqa: TRY003 # FIXME CoP
        f"sorters is a not a SortersDefinition but is a {type(sorters)}"
    )


def _sorter_from_str(sort_key: str) -> Sorter:
    """Convert a list of strings to Sorter objects

    Args:
        sort_key: A batch metadata key which will be used to sort batches on a data asset.
                  This can be prefixed with a + or - to indicate increasing or decreasing
                  sorting.  If not specified, defaults to increasing order.
    """
    if sort_key[0] == "-":
        return Sorter(key=sort_key[1:], reverse=True)

    if sort_key[0] == "+":
        return Sorter(key=sort_key[1:], reverse=False)

    return Sorter(key=sort_key, reverse=False)


# It would be best to bind this to ExecutionEngine, but we can't now due to circular imports
_ExecutionEngineT = TypeVar("_ExecutionEngineT")


DatasourceT = TypeVar("DatasourceT", bound="Datasource")


@public_api
class DataAsset(GenericBaseModel, Generic[DatasourceT, PartitionerT], ABC):
    """
    A Data Asset is a collection of records within a Data Source, which is usually named based
    on the underlying data system and sliced to correspond to a desired specification.

    Data Assets are used to specify how Great Expectations will organize data into Batches.
    """

    # To subclass a DataAsset one must define `type` as a Class literal explicitly on the sublass
    # as well as implementing the methods in the `Abstract Methods` section below.
    # Some examples:
    # * type: Literal["MyAssetTypeID"] = "MyAssetTypeID",
    # * type: Literal["table"] = "table"
    # * type: Literal["csv"] = "csv"
    name: str
    type: str
    id: Optional[uuid.UUID] = Field(default=None, description="DataAsset id")

    # TODO: order_by should no longer be used and should be removed
    order_by: List[Sorter] = Field(default_factory=list)
    batch_metadata: BatchMetadata = pydantic.Field(default_factory=dict)
    batch_definitions: List[BatchDefinition] = Field(default_factory=list)

    # non-field private attributes
    _datasource: DatasourceT = pydantic.PrivateAttr()
    _data_connector: Optional[DataConnector] = pydantic.PrivateAttr(default=None)
    _test_connection_error_message: Optional[str] = pydantic.PrivateAttr(default=None)

    @property
    def datasource(self) -> DatasourceT:
        return self._datasource

    def test_connection(self) -> None:
        """Test the connection for the DataAsset.

        Raises:
            TestConnectionError: If the connection test fails.
        """
        raise NotImplementedError(
            """One needs to implement "test_connection" on a DataAsset subclass."""
        )

    def get_batch_parameters_keys(
        self, partitioner: Optional[PartitionerT] = None
    ) -> tuple[str, ...]:
        raise NotImplementedError(
            """One needs to implement "get_batch_parameters_keys" on a DataAsset subclass."""
        )

    def build_batch_request(
        self,
        options: Optional[BatchParameters] = None,
        batch_slice: Optional[BatchSlice] = None,
        partitioner: Optional[PartitionerT] = None,
    ) -> BatchRequest[PartitionerT]:
        """A batch request that can be used to obtain batches for this DataAsset.

        Args:
            options: A dict that can be used to filter the batch groups returned from the asset.
                The dict structure depends on the asset type. The available keys for dict can be obtained by
                calling get_batch_parameters_keys(...).
            batch_slice: A python slice that can be used to limit the sorted batches by index.
                e.g. `batch_slice = "[-5:]"` will request only the last 5 batches after the options filter is applied.
            partitioner: A Partitioner used to narrow the data returned from the asset.

        Returns:
            A BatchRequest object that can be used to obtain a batch from an asset by calling the
            get_batch method.
        """  # noqa: E501 # FIXME CoP
        raise NotImplementedError(
            """One must implement "build_batch_request" on a DataAsset subclass."""
        )

    @abstractmethod
    def get_batch_identifiers_list(self, batch_request: BatchRequest) -> List[dict]: ...

    @abstractmethod
    def get_batch(self, batch_request: BatchRequest) -> Batch: ...

    def _validate_batch_request(self, batch_request: BatchRequest) -> None:
        """Validates the batch_request has the correct form.

        Args:
            batch_request: A batch request object to be validated.
        """
        raise NotImplementedError(
            """One must implement "_validate_batch_request" on a DataAsset subclass."""
        )

    # End Abstract Methods

    def add_batch_definition(
        self,
        name: str,
        partitioner: Optional[PartitionerT] = None,
    ) -> BatchDefinition[PartitionerT]:
        """Add a BatchDefinition to this DataAsset.
        BatchDefinition names must be unique within a DataAsset.

        If the DataAsset is tied to a DataContext, the BatchDefinition will be persisted.

        Args:
            name (str): Name of the new batch definition.
            partitioner: Optional Partitioner to partition this BatchDefinition

        Returns:
            BatchDefinition: The new batch definition.
        """
        batch_definition_names = {bc.name for bc in self.batch_definitions}
        if name in batch_definition_names:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f'"{name}" already exists (all existing batch_definition names are {", ".join(batch_definition_names)})'  # noqa: E501 # FIXME CoP
            )

        # Let mypy know that self.datasource is a Datasource (it is currently bound to MetaDatasource)  # noqa: E501 # FIXME CoP
        assert isinstance(self.datasource, Datasource)

        batch_definition = BatchDefinition[PartitionerT](name=name, partitioner=partitioner)
        batch_definition.set_data_asset(self)
        self.batch_definitions.append(batch_definition)
        self.update_batch_definition_field_set()
        if self.datasource.data_context:
            try:
                batch_definition = self.datasource.add_batch_definition(batch_definition)
            except Exception:
                self.batch_definitions.remove(batch_definition)
                self.update_batch_definition_field_set()
                raise
        self.update_batch_definition_field_set()
        return batch_definition

    @public_api
    def delete_batch_definition(self, name: str) -> None:
        """Delete a batch definition.

        Args:
            name (str): Name of the BatchDefinition to delete.
        """
        try:
            batch_def = self.get_batch_definition(name)
        except KeyError as err:
            # We collect the names as a list because while we shouldn't have more than 1
            # batch definition with the same name, we want to represent it if it does occur.
            batch_definition_names = [bc.name for bc in self.batch_definitions]
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f'"{name}" does not exist. Existing batch_definition names are {batch_definition_names})'  # noqa: E501 # FIXME CoP
            ) from err
        self._delete_batch_definition(batch_def)

    def _delete_batch_definition(self, batch_definition: BatchDefinition[PartitionerT]) -> None:
        # Let mypy know that self.datasource is a Datasource (it is currently bound to MetaDatasource)  # noqa: E501 # FIXME CoP
        assert isinstance(self.datasource, Datasource)

        self.batch_definitions.remove(batch_definition)
        if self.datasource.data_context:
            try:
                self.datasource.delete_batch_definition(batch_definition)
            except Exception:
                self.batch_definitions.append(batch_definition)
                raise

        self.update_batch_definition_field_set()

    def update_batch_definition_field_set(self) -> None:
        """Ensure that we have __fields_set__ set correctly for batch_definitions to ensure we serialize IFF needed."""  # noqa: E501 # FIXME CoP

        has_batch_definitions = len(self.batch_definitions) > 0
        if "batch_definitions" in self.__fields_set__ and not has_batch_definitions:
            self.__fields_set__.remove("batch_definitions")
        elif "batch_definitions" not in self.__fields_set__ and has_batch_definitions:
            self.__fields_set__.add("batch_definitions")

    @public_api
    def get_batch_definition(self, name: str) -> BatchDefinition[PartitionerT]:
        """Get a batch definition.

        Args:
            name (str): Name of the BatchDefinition to get.
        Raises:
            KeyError: If the BatchDefinition does not exist.
        """
        batch_definitions = [
            batch_definition
            for batch_definition in self.batch_definitions
            if batch_definition.name == name
        ]
        if len(batch_definitions) == 0:
            raise KeyError(  # noqa: TRY003 # FIXME CoP
                f"BatchDefinition {name} not found"
            )
        elif len(batch_definitions) > 1:
            # Our add_batch_definition() method should enforce that different
            # batch definitions do not share a name.
            raise KeyError(  # noqa: TRY003 # FIXME CoP
                f"Multiple keys for {name} found"
            )
        return batch_definitions[0]

    def _batch_parameters_are_valid(
        self, options: BatchParameters, partitioner: Optional[PartitionerT]
    ) -> bool:
        valid_options = self.get_batch_parameters_keys(partitioner=partitioner)
        return set(options.keys()).issubset(set(valid_options))

    @pydantic.validator("batch_metadata", pre=True)
    def ensure_batch_metadata_is_not_none(cls, value: Any) -> Union[dict, Any]:
        """If batch metadata is None, replace it with an empty dict."""
        if value is None:
            return {}
        return value

    def _get_batch_metadata_from_batch_request(
        self, batch_request: BatchRequest, ignore_options: Sequence = ()
    ) -> BatchMetadata:
        """Performs config variable substitution and populates batch parameters for
        Batch.metadata at runtime.
        """
        batch_metadata = copy.deepcopy(self.batch_metadata)
        if not self._datasource.data_context:
            raise MissingDataContextError()
        config_variables = self._datasource.data_context.config_variables
        batch_metadata = _ConfigurationSubstitutor().substitute_all_config_variables(
            data=batch_metadata, replace_variables_dict=config_variables
        )
        batch_metadata.update(
            copy.deepcopy(
                {k: v for k, v in batch_request.options.items() if k not in ignore_options}
            )
        )
        return batch_metadata

    # Sorter methods
    @pydantic.validator("order_by", pre=True)
    def _order_by_validator(
        cls, order_by: Optional[List[Union[Sorter, str, dict]]] = None
    ) -> List[Sorter]:
        if order_by:
            raise DataAssetInitializationError(
                message="'order_by' is no longer a valid argument. "
                "Sorting should be configured in a batch definition."
            )
        return []

    def sort_batches(
        self, batch_list: List[Batch], partitioner: PartitionerSortingProtocol
    ) -> List[Batch]:
        """Sorts batch_list in place in the order configured in this DataAsset.
        Args:
            batch_list: The list of batches to sort in place.
            partitioner: Configuration used to determine sort.
        """

        def get_value(key: str) -> Callable[[Batch], Any]:
            return lambda bd: bd.metadata[key]

        return self._sort_batch_data_list(batch_list, partitioner, get_value)

    def sort_legacy_batch_definitions(
        self,
        legacy_batch_definition_list: List[LegacyBatchDefinition],
        partitioner: PartitionerSortingProtocol,
    ) -> List[LegacyBatchDefinition]:
        """Sorts batch_definition_list in the order configured by the partitioner."""

        def get_value(key: str) -> Callable[[LegacyBatchDefinition], Any]:
            return lambda bd: bd.batch_identifiers[key]

        return self._sort_batch_data_list(legacy_batch_definition_list, partitioner, get_value)

    def sort_batch_identifiers_list(
        self, batch_identfiers_list: List[dict], partitioner: PartitionerSortingProtocol
    ) -> List[dict]:
        """Sorts batch_identfiers_list in the order configured by the partitioner."""

        def get_value(key: str) -> Callable[[dict], Any]:
            return lambda d: d[key]

        return self._sort_batch_data_list(batch_identfiers_list, partitioner, get_value)

    def _sort_batch_data_list(
        self,
        batch_data_list: List[_T],
        partitioner: PartitionerSortingProtocol,
        get_value: Callable[[str], Any],
    ) -> List[_T]:
        """Sorts batch_data_list in the order configured by the partitioner."""
        reverse = not partitioner.sort_ascending
        for key in reversed(partitioner.param_names):
            try:
                batch_data_list = sorted(
                    batch_data_list,
                    key=functools.cmp_to_key(
                        _sort_batch_identifiers_with_none_metadata_values(get_value(key))
                    ),
                    reverse=reverse,
                )
            except KeyError as e:
                raise KeyError(  # noqa: TRY003 # FIXME CoP
                    f"Trying to sort {self.name}'s batches on key {key}, "
                    "which isn't available on all batches."
                ) from e
        return batch_data_list


def _sort_batch_identifiers_with_none_metadata_values(
    get_val: Callable[[_T], Any],
) -> Callable[[_T, _T], int]:
    def _compare_function(a: _T, b: _T) -> int:
        a_val = get_val(a)
        b_val = get_val(b)

        if a_val is not None and b_val is not None:
            if a_val < b_val:
                return -1
            elif a_val > b_val:
                return 1
            else:
                return 0
        elif a_val is None and b_val is None:
            return 0
        elif a_val is None:  # b.metadata_val is not None
            return -1
        else:  # b[key] is None
            return 1

    return _compare_function


# If a Datasource can have more than 1 _DataAssetT, this will need to change.
_DataAssetT = TypeVar("_DataAssetT", bound=DataAsset)


@public_api
class Datasource(
    FluentBaseModel,
    Generic[_DataAssetT, _ExecutionEngineT],
    metaclass=MetaDatasource,
):
    """
    A Datasource provides a standard API for accessing and interacting with data from
    a wide variety of source systems.
    """

    # To subclass Datasource one needs to define:
    # asset_types
    # type
    # assets
    #
    # The important part of defining `assets` is setting the Dict type correctly.
    # In addition, one must define the methods in the `Abstract Methods` section below.
    # If one writes a class level docstring, this will become the documenation for the
    # data context method `data_context.data_sources.add_my_datasource` method.

    # class attrs
    asset_types: ClassVar[Sequence[Type[DataAsset]]] = []
    # Not all Datasources require a DataConnector
    data_connector_type: ClassVar[Optional[Type[DataConnector]]] = None
    # Datasource sublcasses should update this set if the field should not be passed to the execution engine  # noqa: E501 # FIXME CoP
    _EXTRA_EXCLUDED_EXEC_ENG_ARGS: ClassVar[Set[str]] = set()
    _type_lookup: ClassVar[TypeLookup]  # This attribute is set in `MetaDatasource.__new__`
    # Setting this in a Datasource subclass will override the execution engine type.
    # The primary use case is to inject an execution engine for testing.
    execution_engine_override: ClassVar[Optional[Type[_ExecutionEngineT]]] = None  # type: ignore[misc]  # ClassVar cannot contain type variables

    # instance attrs
    type: str
    name: str
    id: Optional[uuid.UUID] = Field(default=None, description="Datasource id")
    assets: MutableSequence[_DataAssetT] = []

    # private attrs
    _data_context: Union[GXDataContext, None] = pydantic.PrivateAttr(None)
    _cached_execution_engine_kwargs: Dict[str, Any] = pydantic.PrivateAttr({})
    _execution_engine: Union[_ExecutionEngineT, None] = pydantic.PrivateAttr(None)

    @property
    def _config_provider(self) -> Union[_ConfigurationProvider, None]:
        return getattr(self._data_context, "config_provider", None)

    @property
    def data_context(self) -> GXDataContext | None:
        """The data context that this datasource belongs to.

        This method should only be used by library implementers.
        """
        return self._data_context

    @pydantic.validator("assets", each_item=True)
    @classmethod
    def _load_asset_subtype(
        cls: Type[Datasource[_DataAssetT, _ExecutionEngineT]], data_asset: DataAsset
    ) -> _DataAssetT:
        """
        Some `data_asset` may be loaded as a less specific asset subtype different than
        what was intended.
        If a more specific subtype is needed the `data_asset` will be converted to a
        more specific `DataAsset`.
        """
        logger.debug(f"Loading '{data_asset.name}' asset ->\n{pf(data_asset, depth=4)}")
        asset_type_name: str = data_asset.type
        asset_type: Type[_DataAssetT] = cls._type_lookup[asset_type_name]

        if asset_type is type(data_asset):
            # asset is already the intended type
            return data_asset

        # strip out asset default kwargs
        kwargs = data_asset.dict(exclude_unset=True)
        logger.debug(f"{asset_type_name} - kwargs\n{pf(kwargs)}")

        cls._update_asset_forward_refs(asset_type)

        asset_of_intended_type = asset_type(**kwargs)
        logger.debug(f"{asset_type_name} - {asset_of_intended_type!r}")
        return asset_of_intended_type

    @pydantic.validator(_ASSETS_KEY, each_item=True)
    def _update_batch_definitions(cls, data_asset: DataAsset) -> DataAsset:
        for batch_definition in data_asset.batch_definitions:
            batch_definition.set_data_asset(data_asset)
        return data_asset

    def _execution_engine_type(self) -> Type[_ExecutionEngineT]:
        """Returns the execution engine to be used"""
        return self.execution_engine_override or self.execution_engine_type

    def add_batch_definition(
        self, batch_definition: BatchDefinition[PartitionerT]
    ) -> BatchDefinition[PartitionerT]:
        asset_name = batch_definition.data_asset.name
        if not self.data_context:
            raise DataContextError(  # noqa: TRY003 # FIXME CoP
                "Cannot save datasource without a data context."
            )

        loaded_datasource = self.data_context.data_sources.get(self.name)
        if loaded_datasource is not self:
            # CachedDatasourceDict will return self; only add batch definition if this is a remote
            # copy
            assert isinstance(loaded_datasource, Datasource)
            loaded_asset = loaded_datasource.get_asset(asset_name)
            loaded_asset.batch_definitions.append(batch_definition)
            loaded_asset.update_batch_definition_field_set()
        updated_datasource = self.data_context.update_datasource(loaded_datasource)
        assert isinstance(updated_datasource, Datasource)

        updated_asset = updated_datasource.get_asset(asset_name)
        updated_batch_definition = updated_asset.get_batch_definition(batch_definition.name)
        if batch_definition is not updated_batch_definition:
            # update in memory copy with the new ID
            batch_definition.id = updated_batch_definition.id
        return updated_batch_definition

    def delete_batch_definition(self, batch_definition: BatchDefinition[PartitionerT]) -> None:
        asset_name = batch_definition.data_asset.name
        if not self.data_context:
            raise DataContextError(  # noqa: TRY003 # FIXME CoP
                "Cannot save datasource without a data context."
            )

        loaded_datasource = self.data_context.data_sources.get(self.name)
        if loaded_datasource is not self:
            # CachedDatasourceDict will return self; only add batch definition if this is a remote
            # copy
            assert isinstance(loaded_datasource, Datasource)
            loaded_asset = loaded_datasource.get_asset(asset_name)
            loaded_asset.batch_definitions.remove(batch_definition)
            loaded_asset.update_batch_definition_field_set()
        updated_datasource = self.data_context.update_datasource(loaded_datasource)
        assert isinstance(updated_datasource, Datasource)

    def get_execution_engine(self) -> _ExecutionEngineT:
        current_execution_engine_kwargs = self.dict(
            exclude=self._get_exec_engine_excludes(),
            config_provider=self._config_provider,
        )
        if (
            current_execution_engine_kwargs != self._cached_execution_engine_kwargs
            or not self._execution_engine
        ):
            self._execution_engine = self._execution_engine_type()(
                **current_execution_engine_kwargs
            )
            self._cached_execution_engine_kwargs = current_execution_engine_kwargs
        return self._execution_engine

    def get_batch(self, batch_request: BatchRequest) -> Batch:
        """A Batch that corresponds to the BatchRequest.

        Args:
            batch_request: A batch request for this asset. Usually obtained by calling
                build_batch_request on the asset.

        Returns:
            A Batch that matches the options specified in the batch request.
        """
        data_asset = self.get_asset(batch_request.data_asset_name)
        return data_asset.get_batch(batch_request)

    def get_batch_identifiers_list(self, batch_request: BatchRequest) -> List[dict]:
        data_asset = self.get_asset(batch_request.data_asset_name)
        return data_asset.get_batch_identifiers_list(batch_request)

    def get_assets_as_dict(self) -> MutableMapping[str, _DataAssetT]:
        """Returns available DataAsset objects as dictionary, with corresponding name as key.

        Returns:
            Dictionary of "_DataAssetT" objects with "name" attribute serving as key.
        """
        asset: _DataAssetT
        assets_as_dict: MutableMapping[str, _DataAssetT] = {
            asset.name: asset for asset in self.assets
        }

        return assets_as_dict

    def get_asset_names(self) -> Set[str]:
        """Returns the set of available DataAsset names

        Returns:
            Set of available DataAsset names.
        """
        asset: _DataAssetT
        return {asset.name for asset in self.assets}

    @public_api
    def get_asset(self, name: str) -> _DataAssetT:
        """Returns the DataAsset referred to by asset_name

        Args:
            name: name of DataAsset sought.

        Returns:
            if named "DataAsset" object exists; otherwise, exception is raised.
        """
        # This default implementation will be used if protocol is inherited
        try:
            asset: _DataAssetT
            found_asset: _DataAssetT = list(filter(lambda asset: asset.name == name, self.assets))[
                0
            ]
            found_asset._datasource = self
            return found_asset
        except IndexError as exc:
            raise LookupError(  # noqa: TRY003 # FIXME CoP
                f'"{name}" not found. Available assets are ({", ".join(self.get_asset_names())})'
            ) from exc

    @public_api
    def delete_asset(self, name: str) -> None:
        """Removes the DataAsset referred to by asset_name from internal list of available DataAsset objects.

        Args:
            name: name of DataAsset to be deleted.
        """  # noqa: E501 # FIXME CoP
        from great_expectations.data_context import CloudDataContext

        asset: _DataAssetT
        asset = self.get_asset(name=name)

        if self._data_context and isinstance(self._data_context, CloudDataContext):
            self._data_context._delete_asset(id=str(asset.id))

        self.assets = list(filter(lambda asset: asset.name != name, self.assets))
        self._save_context_project_config()

    def _add_asset(self, asset: _DataAssetT, connect_options: dict | None = None) -> _DataAssetT:
        """Adds an asset to a datasource

        Args:
            asset: The DataAsset to be added to this datasource.
        """
        # The setter for datasource is non-functional, so we access _datasource directly.
        # See the comment in DataAsset for more information.
        asset._datasource = self

        if not connect_options:
            connect_options = {}
        self._build_data_connector(asset, **connect_options)

        asset.test_connection()

        asset_names: Set[str] = self.get_asset_names()
        if asset.name in asset_names:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f'"{asset.name}" already exists (all existing assets are {", ".join(asset_names)})'
            )

        self.assets.append(asset)

        # if asset was added to a cloud FDS, _update_fluent_datasource will return FDS fetched from cloud,  # noqa: E501 # FIXME CoP
        # which will contain the new asset populated with an id
        if self._data_context:
            updated_datasource = self._data_context._update_fluent_datasource(datasource=self)
            assert isinstance(updated_datasource, Datasource)
            if asset_id := updated_datasource.get_asset(name=asset.name).id:
                asset.id = asset_id

        return asset

    def _save_context_project_config(self) -> None:
        """Check if a DataContext is available and save the project config."""
        if self._data_context:
            try:
                self._data_context._save_project_config()
            except TypeError as type_err:
                warnings.warn(str(type_err), GxSerializationWarning)

    def _rebuild_asset_data_connectors(self) -> None:
        """
        If Datasource required a data_connector we need to build the data_connector for each asset.

        A warning is raised if a data_connector cannot be built for an asset.
        Not all users will have access to the needed dependencies (packages or credentials) for every asset.
        Missing dependencies will stop them from using the asset but should not stop them from loading it from config.
        """  # noqa: E501 # FIXME CoP
        asset_build_failure_direct_cause: dict[str, Exception | BaseException] = {}

        if self.data_connector_type:
            for data_asset in self.assets:
                try:
                    # check if data_connector exist before rebuilding?
                    connect_options = getattr(data_asset, "connect_options", {})
                    self._build_data_connector(data_asset, **connect_options)
                except Exception as dc_build_err:
                    logger.info(
                        f"Unable to build data_connector for {self.type} {data_asset.type} {data_asset.name}",  # noqa: E501 # FIXME CoP
                        exc_info=True,
                    )
                    # reveal direct cause instead of generic, unhelpful MyDatasourceError
                    asset_build_failure_direct_cause[data_asset.name] = (
                        dc_build_err.__cause__ or dc_build_err
                    )
        if asset_build_failure_direct_cause:
            # TODO: allow users to opt out of these warnings
            names_and_error: List[str] = [
                f"{name}:{type(exc).__name__}"
                for (name, exc) in asset_build_failure_direct_cause.items()
            ]
            warnings.warn(
                f"data_connector build failure for {self.name} assets - {', '.join(names_and_error)}",  # noqa: E501 # FIXME CoP
                category=RuntimeWarning,
            )

    @staticmethod
    def _update_asset_forward_refs(asset_type: Type[_DataAssetT]) -> None:
        """Update forward refs of an asset_type if necessary.

        Note, this should be overridden in child datasource classes if forward
        refs need to be updated. For example, in Spark datasources we need to
        update forward refs only if the optional spark dependencies are installed
        so this method is overridden. Here it is a no op.

        Args:
            asset_type: Asset type to update forward refs.

        Returns:
            None, asset refs is updated in place.
        """
        pass

    # Abstract Methods
    @property
    def execution_engine_type(self) -> Type[_ExecutionEngineT]:
        """Return the ExecutionEngine type use for this Datasource"""
        raise NotImplementedError(
            """One needs to implement "execution_engine_type" on a Datasource subclass."""
        )

    def test_connection(self, test_assets: bool = True) -> None:
        """Test the connection for the Datasource.

        Args:
            test_assets: If assets have been passed to the Datasource, an attempt can be made to test them as well.

        Raises:
            TestConnectionError: If the connection test fails.
        """  # noqa: E501 # FIXME CoP
        raise NotImplementedError(
            """One needs to implement "test_connection" on a Datasource subclass."""
        )

    def _build_data_connector(self, data_asset: _DataAssetT, **kwargs) -> None:
        """Any Datasource subclass that utilizes DataConnector should overwrite this method.

        Specific implementations instantiate appropriate DataConnector class and set "self._data_connector" to it.

        Args:
            data_asset: DataAsset using this DataConnector instance
            kwargs: Extra keyword arguments allow specification of arguments used by particular DataConnector subclasses
        """  # noqa: E501 # FIXME CoP
        pass

    @classmethod
    def _get_exec_engine_excludes(cls) -> Set[str]:
        """
        Return a set of field names to exclude from the execution engine.

        All datasource fields are passed to the execution engine by default unless they are in this set.

        Default implementation is to return the combined set of field names from `_EXTRA_EXCLUDED_EXEC_ENG_ARGS`
        and `_BASE_DATASOURCE_FIELD_NAMES`.
        """  # noqa: E501 # FIXME CoP
        return cls._EXTRA_EXCLUDED_EXEC_ENG_ARGS.union(_BASE_DATASOURCE_FIELD_NAMES)

    # End Abstract Methods


# This is used to prevent passing things like `type`, `assets` etc. to the execution engine
_BASE_DATASOURCE_FIELD_NAMES: Final[Set[str]] = {name for name in Datasource.__fields__}


@dataclasses.dataclass(frozen=True)
class HeadData:
    """
    An immutable wrapper around pd.DataFrame for .head() methods which
        are intended to be used for visual inspection of BatchData.
    """

    data: pd.DataFrame

    @override
    def __repr__(self) -> str:
        return self.data.__repr__()


@public_api
class Batch:
    """This represents a batch of data.

    This is usually not the data itself but a hook to the data on an external datastore such as
    a spark or a sql database. An exception exists for pandas or any in-memory datastore.
    """

    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        datasource: Datasource,
        data_asset: DataAsset,
        batch_request: BatchRequest,
        data: BatchData,
        batch_markers: BatchMarkers,
        batch_spec: BatchSpec,
        batch_definition: LegacyBatchDefinition,
        metadata: Dict[str, Any] | None = None,
    ):
        # Immutable attributes
        self._datasource = datasource
        self._data_asset = data_asset
        self._batch_request = batch_request
        self._data = data

        # Immutable legacy attributes
        # TODO: These legacy fields are required but we should figure out how to delete them
        self._batch_markers = batch_markers
        self._batch_spec = batch_spec
        self._batch_definition = batch_definition

        # Mutable Attribute
        # metadata is any arbitrary data one wants to associate with a batch. GX will add arbitrary metadata  # noqa: E501 # FIXME CoP
        # to a batch so developers may want to namespace any custom metadata they add.
        self.metadata = metadata or {}

        # Immutable generated attribute
        self._id = self._create_id()

    def _create_id(self) -> str:
        options_list = []
        for key, value in self.batch_request.options.items():
            if key not in ("path", "dataframe"):
                options_list.append(f"{key}_{value}")
        return "-".join([self.datasource.name, self.data_asset.name, *options_list])

    @property
    def datasource(self) -> Datasource:
        return self._datasource

    @property
    def data_asset(self) -> DataAsset:
        return self._data_asset

    @property
    def batch_request(self) -> BatchRequest:
        return self._batch_request

    @property
    def data(self) -> BatchData:
        return self._data

    @property
    def batch_markers(self) -> BatchMarkers:
        return self._batch_markers

    @property
    def batch_spec(self) -> BatchSpec:
        return self._batch_spec

    @property
    def batch_definition(self) -> LegacyBatchDefinition:
        return self._batch_definition

    @property
    def id(self) -> str:
        return self._id

    def _get_metrics_calculator(self) -> MetricsCalculator:
        self.data.execution_engine.batch_manager.load_batch_list(batch_list=[self])
        return MetricsCalculator(
            execution_engine=self.data.execution_engine,
            show_progress_bars=True,
        )

    @public_api
    @validate_arguments
    def columns(self) -> List[str]:
        """Return column names of this Batch.

        Returns:
            list of column names.
        """
        return self._get_metrics_calculator().columns()

    @public_api
    @validate_arguments
    def head(
        self,
        n_rows: StrictInt = 5,
        fetch_all: StrictBool = False,
    ) -> HeadData:
        """Return the first n rows of this Batch.

        This method returns the first n rows for the Batch based on position.

        For negative values of n_rows, this method returns all rows except the last n rows.

        If n_rows is larger than the number of rows, this method returns all rows.

        Args:
            n_rows: The number of rows to return from the Batch.
            fetch_all: If True, ignore n_rows and return the entire Batch.

        Returns:
            HeadData
        """
        metrics_calculator = self._get_metrics_calculator()
        table_head_df: pd.DataFrame = metrics_calculator.head(
            n_rows=n_rows,
            domain_kwargs={"batch_id": self.id},
            fetch_all=fetch_all,
        )
        return HeadData(data=table_head_df.reset_index(drop=True, inplace=False))

    @overload
    def validate(
        self,
        expect: Expectation,
        *,
        result_format: ResultFormatUnion = DEFAULT_RESULT_FORMAT,
        expectation_parameters: Optional[SuiteParameterDict] = None,
    ) -> ExpectationValidationResult: ...

    @overload
    def validate(
        self,
        expect: ExpectationSuite,
        *,
        result_format: ResultFormatUnion = DEFAULT_RESULT_FORMAT,
        expectation_parameters: Optional[SuiteParameterDict] = None,
    ) -> ExpectationSuiteValidationResult: ...

    @public_api
    def validate(
        self,
        expect: Expectation | ExpectationSuite,
        *,
        result_format: ResultFormatUnion = DEFAULT_RESULT_FORMAT,
        expectation_parameters: Optional[SuiteParameterDict] = None,
    ) -> ExpectationValidationResult | ExpectationSuiteValidationResult:
        """
        Validate the Batch using the provided Expectation or Expectation Suite.

        Args:
            expect: The Expectation or Expectation Suite to validate.
            result_format: The format to return the validation results in.
            expectation_parameters: A dictionary of parameters values for any
                                    expectations using parameterized values (the $PARAMETER syntax).
                                    The keys are the parameter names and the values are the values
                                    to be used for this validation run.

        Returns:
            An ExpectationValidationResult or ExpectationSuiteValidationResult object.

        Raises:
            ValueError: If the expect argument is not an Expectation or an ExpectationSuite.
        """
        from great_expectations.core import ExpectationSuite
        from great_expectations.expectations.expectation import Expectation

        if isinstance(expect, Expectation):
            return self._validate_expectation(
                expect, result_format=result_format, expectation_parameters=expectation_parameters
            )
        elif isinstance(expect, ExpectationSuite):
            return self._validate_expectation_suite(
                expect, result_format=result_format, expectation_parameters=expectation_parameters
            )
        else:
            # If we are type checking, we should never fall through to this case. However, exploratory  # noqa: E501 # FIXME CoP
            # workflows are not being type checked.
            raise ValueError(  # noqa: TRY003, TRY004 # FIXME CoP
                f"Trying to validate something that isn't an Expectation or an ExpectationSuite: {expect}"  # noqa: E501 # FIXME CoP
            )

    def _validate_expectation(
        self,
        expect: Expectation,
        result_format: ResultFormatUnion,
        expectation_parameters: Optional[SuiteParameterDict] = None,
    ) -> ExpectationValidationResult:
        return self._create_validator(
            result_format=result_format,
        ).validate_expectation(expectation=expect, expectation_parameters=expectation_parameters)

    def _validate_expectation_suite(
        self,
        expect: ExpectationSuite,
        result_format: ResultFormatUnion,
        expectation_parameters: Optional[SuiteParameterDict] = None,
    ) -> ExpectationSuiteValidationResult:
        return self._create_validator(
            result_format=result_format,
        ).validate_expectation_suite(
            expectation_suite=expect, expectation_parameters=expectation_parameters
        )

    def _create_validator(self, *, result_format: ResultFormatUnion) -> V1Validator:
        from great_expectations.validator.v1_validator import Validator as V1Validator

        context = self.datasource.data_context
        if context is None:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                "We can't validate batches that are attached to datasources without a data context"
            )

        # note: batch definition is created but NOT added to the asset, as it should not persist
        batch_definition = BatchDefinition(
            name="-".join([self.datasource.name, self.data_asset.name, str(uuid.uuid4())]),
            partitioner=self.batch_request.partitioner,
        )
        batch_definition.set_data_asset(self.data_asset)

        return V1Validator(
            batch_definition=batch_definition,
            batch_parameters=self.batch_request.options,
            result_format=result_format,
        )

    @overload
    def compute_metrics(
        self, metrics: Metric[_MetricResultT]
    ) -> _MetricResultT | MetricErrorResult: ...

    @overload
    def compute_metrics(self, metrics: list[Metric]) -> list[MetricResult]: ...

    def compute_metrics(
        self, metrics: Metric[_MetricResultT] | list[Metric]
    ) -> _MetricResultT | MetricErrorResult | list[MetricResult]:
        """Compute one or more metrics on this Batch.

        Args:
            metrics: A single Metric or list of Metrics to compute.
                     Each Metric must be an instance of a concrete Metric subclass.

        Returns:
            If a single Metric is provided, returns a single MetricResult.
            If a list of Metrics is provided, returns a list of MetricResults,
            in the same order as the input metrics were received.
            For metrics without a defined MetricResult generic type,
            the base MetricResult class will be returned.
            For metrics that raise an exception, a MetricErrorResult will be returned.

        Examples:
            >>> batch.compute_metrics(BatchRowCount())
            BatchRowCountResult(id=..., value=1000)

            >>> batch.compute_metrics([
            ...     BatchRowCount(),
            ...     ColumnMax(column="age")
            ... ])
            [BatchRowCountResult(id=..., value=1000), ColumnMaxResult(id=..., value=85)]

        Notes:
            Until this mypy bug is resolved, lists of Metrics are being incorrectly
            inferred by the static type checker as list[Domain]. You can work around
            this by adding an explicit annotation (e.g. metrics: list[Metric] = ...)
            https://github.com/python/mypy/issues/18712
        """
        is_single_metric = False
        if isinstance(metrics, Metric):
            metrics = [metrics]
            is_single_metric = True

        metrics_calculator = self._get_metrics_calculator()
        metrics_calculator_results, metrics_calculator_errors = metrics_calculator.compute_metrics(
            metric_configurations=[metric.config(batch_id=self.id) for metric in metrics],
            runtime_configuration=None,
        )

        results = []
        for metric in metrics:
            metric_id_for_batch = metric.metric_id_for_batch(batch_id=self.id)
            if metric_id_for_batch in metrics_calculator_results:
                MetricType = MetaMetric.get_registered_metric_class_from_metric_name(metric.name)
                MetricResultType = MetricType.get_metric_result_type()
                value = self._parse_metric_value(
                    metric_name=metric.name,
                    metric_calculator_result=metrics_calculator_results[metric_id_for_batch],
                )
                results.append(MetricResultType(id=metric_id_for_batch, value=value))
            elif metric_id_for_batch in metrics_calculator_errors:
                metrics_calculator_error = metrics_calculator_errors[metric_id_for_batch]
                value = MetricErrorResultValue(
                    exception_message=metrics_calculator_error["exception_info"].exception_message,
                    exception_traceback=metrics_calculator_error[
                        "exception_info"
                    ].exception_traceback,
                )
                results.append(MetricErrorResult(id=metric_id_for_batch, value=value))
            else:
                results.append(
                    MetricErrorResult(
                        id=metric_id_for_batch,
                        value=MetricErrorResultValue(
                            exception_message=f"Metric {metric.name} not found in results: "
                            f"{list(metrics_calculator_results.keys())} or errors: "
                            f"{list(metrics_calculator_errors.keys())}",
                        ),
                    )
                )

        if is_single_metric:
            return results[0]
        return results

    def _parse_metric_value(
        self, metric_name: str, metric_calculator_result: MetricValue
    ) -> MetricValue:
        if metric_name.endswith(MetricNameSuffix.CONDITION) and isinstance(
            metric_calculator_result, tuple
        ):
            value = metric_calculator_result[0]
        else:
            value = metric_calculator_result
        return value
